import logging
import operator
import re
from copy import deepcopy

from .evaluate import ConditionBlock


log = logging.getLogger(__name__)


class TraceableDict(dict):

    """
    Python dict-based class that tracks any changes.
    Format:
        {"action": "add", "key": <key>, "value": <new-value>},
        {"action": "remove", "key": <key>, "value": <old-value>},
        {"action": "update", "key": <key>, "old_value": <old-value>,
                                           "new_value": <new-value>},
    """

    def __init__(self, dict2):
        super().__init__(deepcopy(dict2))
        self._changes = []

    def __setitem__(self, key, value):
        if key not in self:
            self._changes.append({
                'action': 'add',
                'key': key,
                'value': deepcopy(value)
            })
        elif self[key] != value:
            self._changes.append({
                'action': 'update',
                'key': key,
                'old_value': deepcopy(self[key]),
                'new_value': deepcopy(value)
            })
        super().__setitem__(key, value)

    def __delitem__(self, key):
        self._changes.append({
            'action': 'remove',
            'key': key,
            'value': deepcopy(self[key])
        })
        super().__delitem__(key)

    def update(self, dict2=None, **kwargs):
        for key in dict2:
            if key not in self:
                self._changes.append({
                    'action': 'add',
                    'key': key,
                    'value': deepcopy(dict2[key])
                })
            elif self[key] != dict2[key]:
                self._changes.append({
                    'action': 'update',
                    'key': key,
                    'old_value': deepcopy(self[key]),
                    'new_value': deepcopy(dict2[key])
                })
        super().update(dict2, **kwargs)

    @property
    def changes(self):
        return self._changes


# Inspired from https://github.com/faif/python-patterns/blob/master/registry.py
class _RegisteredRule(type):
    """
    This metaclass is designed to register automatically public rule classes
    that inherit from `_Rules()`.
    """
    _registry = {}

    def __new__(mcs, name, bases, attrs):
        cls = type.__new__(mcs, name, bases, attrs)
        name = cls.TYPENAME or cls.__name__.lower()
        if not name.startswith('_'):
            cls.TYPENAME = name
            mcs._registry[name] = cls
        return cls

    @classmethod
    def get(mcs, rtype):
        return mcs._registry[rtype]


class Converter(object):

    """
    A sequence of `_Rule` objects intended to be applied on a dict.
    """

    def __init__(self, rules=None):
        self._rules = rules or []

    @property
    def rules(self):
        return self._rules

    @classmethod
    def from_dict(cls, config):
        """
        Create a `Converter` object from dict 'config'. The dict must look like
        the following:
        {
            "rules": [
                {"type": <rule-type-name>, "fieldname": <name>, ...},
                {"type": <rule-type-name>, "fieldname": <name>, ...},
                {"type": "condition-block", "conditions": [...]},
                {"type": <rule-type-name>, "fieldname": <name>, ...},
            ]
        }
        """
        rules = []
        for rule in config['rules']:
            params = rule.copy()
            rtype = params.pop('type')
            rule_cls = _RegisteredRule.get(rtype)
            rules.append(rule_cls(**params))
        return cls(rules=rules)

    def apply(self, data):
        rules = []
        errors = False
        for rule in self.rules:
            diff = rule.apply(data)
            rules.append(diff)
            if diff is not None and 'error' in diff:
                errors = True

        return {'rules': rules, 'errors': errors}


class FactoryConditionBlock(ConditionBlock, metaclass=_RegisteredRule):

    TYPENAME = 'condition-block'

    def __init__(self, conditions):
        super().__init__(conditions)
        self._changes = {'type': self.TYPENAME, 'conditions': []}

    def condition_validated(self, rules, data):
        """
        Apply rules on data upon validating a condition.
        """
        diff = Converter.from_dict({'rules': rules}).apply(data)
        self._changes['conditions'] = diff['rules']

    def apply(self, data):
        super().apply(data)
        return self._changes


class _Rule(metaclass=_RegisteredRule):

    """
    A rule extracts an entry from a dict , performs an operation on the dict
    and updates the dict with the result of that operation (in-place update).
    Only one field from the input dict can be processed within a rule.
    """

    # Public subclasses of `_Rule` will have a TYPENAME attr set to classname
    # (lowercase) if not set in the subclass itself.
    TYPENAME = None

    def __init__(self, fieldname, *args, **kwargs):
        self.fieldname = fieldname
        self._configure(*args, **kwargs)

    def _configure(self, *args, **kwargs):
        """
        Configure the rule to be applied (e.g. compile a regexp) so that
        everything is ready at runtime to apply it.
        """
        raise NotImplementedError

    @staticmethod
    def track_changes(func):
        """
        Decorator for `Rule.apply(<data>)` methods that return a JSON-formatted
        diff of the changes made by the method itself.
        """
        def wrapper(self, data):
            # Handle data through a traceable dict
            tracker = TraceableDict(data)
            try:
                func(self, tracker)
            except RegexpRuleError as exc:
                return {
                    'type': self.TYPENAME, 'changes': tracker.changes,
                    'error': 'regexp_rule_error', 'error_details': str(exc)
                }
            except ArithmeticRuleError as exc:
                return {
                    'type': self.TYPENAME, 'changes': tracker.changes,
                    'error': 'arithmetic_rule_error', 'error_details': str(exc)
                }
            except UnionRuleError as exc:
                return {
                    'type': self.TYPENAME, 'changes': tracker.changes,
                    'error': 'union_rule_error', 'error_details': str(exc)
                }
            # We want to keep the object reference
            data.clear()
            data.update(tracker)
            return {'type': self.TYPENAME, 'changes': tracker.changes}

        return wrapper

    def apply(self, data):
        """
        Execute an operation on one field of the dict `data` and returns an
        diff.
        """
        raise NotImplementedError


class _RegexpRule(_Rule):

    """
    Define a regular expression that shall be applied through several kinds of
    operations to a string.
    If the regexp pattern has named capturing parenthesis it is used to update
    the `data` dict passed to `apply()`. If a group from the match object and
    a key from `data` have the same name, the group overrides the existing
    dict value.
    """

    def _configure(self, pattern, flags=0):
        self.regexp = re.compile(pattern, flags=flags)

    def _run_regexp(self, string):
        raise NotImplementedError

    @_Rule.track_changes
    def apply(self, data):
        try:
            string = data[self.fieldname]
            assert string is not None
        except (KeyError, AssertionError):
            # No data to process
            log.debug('Regexp : unknown or None field %s, ignoring', self.fieldname)
        else:
            resdict = self._run_regexp(string)
            data.update(resdict)


class RegexpRuleError(TypeError):
    pass


class Extract(_RegexpRule):

    """
    Scan through a string looking for the first location where the regular
    expression pattern produces a match. What matters is the resulting dict of
    captured substrings.
    """

    def _configure(self, pattern, flags=0, pos=None, endpos=None):
        super()._configure(pattern, flags=flags)
        self._pos_args = tuple(f for f in (pos, endpos) if f is not None)

    def _run_regexp(self, string):
        if not self.regexp.groupindex:
            raise RegexpRuleError("regex is invalid, ensure a group is captured")
        args = (string,) + self._pos_args
        match = self.regexp.search(*args)
        if match is not None:
            return match.groupdict()
        else:
            return {}


class Sub(_RegexpRule):

    """
    Build a string obtained by replacing the leftmost non-overlapping
    occurrences of regexp pattern in fieldname from `data` by a replacement
    string.
    """

    def _configure(self, pattern, repl, flags=0, count=0):
        super()._configure(pattern, flags=flags)
        self.repl, self.count = repl, count

    def _run_regexp(self, string):
        res = self.regexp.sub(self.repl, string, count=self.count)
        return {self.fieldname: res}


class Set(_Rule):

    """
    Set or update a field in the `data` dict.
    """

    def _configure(self, value):
        self.value = value

    @_Rule.track_changes
    def apply(self, data):
        data[self.fieldname] = self.value


class Copy(_Rule):

    """
    Copy a data into another field.
    """

    def _configure(self, copy):
        self.copy = copy

    @_Rule.track_changes
    def apply(self, data):
        try:
            data[self.copy] = data[self.fieldname]
        except KeyError:
            log.debug('Copy : unknown field %s, ignoring', self.fieldname)


class Unset(_Rule):

    """
    Remove a field from the `data` dict.
    """

    def _configure(self):
        pass

    @_Rule.track_changes
    def apply(self, data):
        try:
            del data[self.fieldname]
        except KeyError:
            log.debug('Unset : unknown field %s, ignoring', self.fieldname)


class Lookup(_Rule):

    """
    Implements a dead-simple lookup table which perform case sensitive
    (default) or insensitive (icase=True) lookups.
    """

    def _configure(self, table=None, icase=False):
        self.table = table or dict()
        self.icase = icase
        if icase:
            self.table = {k.lower(): v for k, v in table.items()}

    @_Rule.track_changes
    def apply(self, data):
        """
        The 1st entry in the lookup table that matches the string from `data`
        replaces it. Note that entries in the lookup table are evaluated in an
        unpredictable order.
        """
        try:
            fieldval = str(data[self.fieldname])
            if self.icase:
                fieldval = fieldval.lower()
            data[self.fieldname] = self.table[fieldval]
        except KeyError as err:
            log.debug("Lookup: fieldname '%s' not in data, ignoring", err)
            return


class Lower(_Rule):

    """
    Lower case a string.
    """

    def _configure(self):
        pass

    @_Rule.track_changes
    def apply(self, data):
        try:
            fieldval = data[self.fieldname]
            data[self.fieldname] = fieldval.lower()
        except KeyError as err:
            log.debug("Lower: fieldname '%s' not in data, ignoring", err)
        except AttributeError as err:
            log.debug("Upper: fieldname '%s' invalid, ignoring", err)


class Upper(_Rule):

    """
    Upper case a string.
    """

    def _configure(self):
        pass

    @_Rule.track_changes
    def apply(self, data):
        try:
            fieldval = data[self.fieldname]
            data[self.fieldname] = fieldval.upper()
        except KeyError as err:
            log.debug("Upper: fieldname '%s' not in data, ignoring", err)
        except AttributeError as err:
            log.debug("Upper: fieldname '%s' invalid, ignoring", err)


class ArithmeticRuleError(Exception):
    pass


class Arithmetic(_Rule):

    """
    Arithmetic rule to add, substrack, multiply and divide fields.
    """

    # List available operators and their associated types.
    OPS = {
        '+': (operator.add, {
            str: (str,),
            int: (int, float),
            float: (int, float),
        }),
        '-': (operator.sub, {
            int: (int, float),
            float: (int, float),
        }),
        '*': (operator.mul, {
            int: (int, float),
            float: (int, float),
        }),
        '/': (operator.truediv, {
            int: (int, float),
            float: (int, float),
        }),
        '%': (operator.mod, {
            int: (int, float),
            float: (int, float),
        }),
    }

    def _configure(self, operator, operand1, operand2):
        self.op, self.types = self.OPS[operator]
        self.operands = (operand1, operand2)

    def _compute_operands(self, data):
        operands = tuple()
        # We replace placeholders with the actual data
        for op in self.operands:
            if isinstance(op, str) and re.match(r'^@[\w-]+$', op):
                op = data[op.split('@')[1]]
            operands += (op,)
        return operands

    @_Rule.track_changes
    def apply(self, data):
        try:
            operand1, operand2 = self._compute_operands(data)
        except KeyError as exc:
            log.debug('Unknown key %s for arithmetic rule', exc)
            raise ArithmeticRuleError(exc)
        except ValueError as exc:
            log.debug('Unusable operands: %s (%s)', exc, exc.__class__)
            raise ArithmeticRuleError(exc)

        type1 = type(operand1)
        type2 = type(operand2)
        if type1 not in self.types or type2 not in self.types[type1]:
            raise ArithmeticRuleError(
                'Bad operand types ({} against {})'.format(type1, type2)
            )

        try:
            result = self.op(operand1, operand2)
        except TypeError as exc:
            log.debug(exc)
            return

        if isinstance(result, float):
            # Arbitrary 3-round value
            result = round(result, 3)

        data[self.fieldname] = result


class UnionRuleError(Exception):
    pass


class Union(_Rule):

    """
    Union rule to merge a list or a dict to another.
    """

    def _configure(self, operand1, operand2):
        self.operands = (operand1, operand2)

    def _compute_operands(self, data):
        computed = tuple()
        for op in self.operands:
            if isinstance(op, str) and re.match(r'^@[\w-]+$', op):
                computed += (data[op.split('@')[1]],)
            else:
                computed += (op,)
        return computed

    def _union(self, a, b):
        if isinstance(a, dict) and isinstance(b, dict):
            return {**a, **b}
        elif isinstance(a, list) and isinstance(b, list):
            return a + [item for item in b if item not in a]
        raise UnionRuleError('union available for two dicts or lists')

    @_Rule.track_changes
    def apply(self, data):
        try:
            operand1, operand2 = self._compute_operands(data)
        except KeyError as exc:
            log.debug('Unknown key %s for arithmetic rule', exc)
            return

        try:
            data[self.fieldname] = self._union(operand1, operand2)
        except TypeError as exc:
            log.debug(exc)



import logging
from tukio.task import register
from tukio.task.holder import TaskHolder


log = logging.getLogger(__name__)


@register('report', 'execute')
class ReportTask(TaskHolder):

    """
    An base task that can perform some http calls on an API.
    """

    NAME = 'report'

    async def execute(self, event):
        log.info('workflow report data: \033[92m"{}"\033[1;m'.format(event.data))
        return event.data


import re
import ast
from collections import defaultdict
import logging


log = logging.getLogger(__name__)


EXPRESSIONS = [
    # Types of values
    ast.Dict,
    ast.List,
    ast.NameConstant,
    ast.Num,
    ast.Set,
    ast.Str,
    ast.Tuple,
    # Types of operations
    ast.Compare,
    ast.BoolOp,
    ast.UnaryOp
]

OPERATORS = [
    ast.And,
    ast.Eq, ast.NotEq,
    ast.Lt, ast.LtE,
    ast.Gt, ast.GtE,
    ast.In, ast.NotIn,
    ast.Invert,
    ast.Is, ast.IsNot,
    ast.Not,
    ast.Or,
    ast.UAdd,
    ast.USub
]

CONTEXTS = [ast.Load]

AUTHORIZED_TYPES = EXPRESSIONS + OPERATORS + CONTEXTS


def safe_eval(expr):
    """
    Ensures an expression only defines authorized operations (no call to
    functions, no variable assignement...) and evaluates it.
    """
    tree = ast.parse(expr, mode='eval').body
    for node in ast.walk(tree):
        if not type(node) in AUTHORIZED_TYPES:
            raise TypeError("forbidden type {} found in {}".format(node, expr))
    return bool(eval(expr))


class ConditionBlock:

    def __init__(self, conditions):
        # Check there is at least one condition
        if len(conditions) == 0:
            raise ValueError('no condition in condition block')
        # Check first condition is 'if'
        if conditions[0]['type'] != 'if':
            raise TypeError("first condition must be an 'if'")
        # Check next conditions (if any)
        if len(conditions) >= 2:
            for cond in conditions[1:-1]:
                # All intermediate conditions must be 'elif'
                if cond['type'] != 'elif':
                    raise TypeError("expected 'elif' condition,"
                                    " got '{}'".format(cond))
            # The last condition can be either an 'elif' or an 'else'
            if conditions[-1]['type'] not in ('elif', 'else'):
                raise TypeError("last condition must be 'elif' or 'else',"
                                " got '{}'".format(conditions[-1]))
        self._conditions = conditions

    def _clean_condition(self, condition, data):
        """
        Format the condition string (as eval-compliant code).
        nb: variable replacement should be `@variable_name` formatted.
        """
        # /!\ This regex forbids the use of ' and " in a string
        # See https://regex101.com/r/hUueag/7
        match = re.findall(
            r' *(and|or)? *\( *(@\S*|None|True|False|[\"\'\[][^\'\"]*[\'\"\]]|\d+) +([=<>!]=?|not in|in|not) +(@\S*|None|True|False|[\"\'\[][^\'\"]*[\'\"\]]|\d+) *\) *',
            condition
        )
        if not match:
            return condition

        def replace(match):
            key = match.group('var_name')
            value = data.get(key)
            placeholder = '{!r}' if isinstance(value, str) else '{}'
            return placeholder.format(value)

        # Reconstruct a cleaned string from the operation parts.
        cleaned = ''
        for operation in match:
            # Get 'and' or 'or' operation
            andor = operation[0]
            # Restructure condition string, striping any trailing space
            ops = []
            ops.append(re.sub(r'^@(?P<var_name>\w+)$', replace, operation[1]))
            ops.append(operation[2])
            ops.append(re.sub(r'^@(?P<var_name>\w+)$', replace, operation[3]))
            cleaned += '{}({})'.format(andor, ' '.join(ops))
        return cleaned

    def condition_validated(self, condition, data):
        """
        To be overridden to do your own logic once the condition has
        been validated (True) by `self._evaluate`.
        """
        raise NotImplementedError

    def apply(self, data):
        """
        Iterate through the conditions and stop at first validated condition.
        """
        for condition in self._conditions:
            # If type 'else', set given next tasks and leave
            if condition['type'] == 'else':
                self.condition_validated(condition['rules'], data)
                return
            # Else find the condition and evaluate it
            cleaned = self._clean_condition(condition['condition'], data)
            log.debug('arithmetics: trying %s', cleaned)
            if safe_eval(cleaned):
                log.debug(
                    'arithmetics: validated condition "%s" as "%s"',
                    condition, cleaned
                )
                self.condition_validated(condition['rules'], data)
                return



import asyncio
import logging
from socket import error as SocketError
from aioredis import create_reconnecting_redis, RedisError

from nyuki.services import Service


log = logging.getLogger(__name__)


def memsafe(coro):
    async def wrapper(*args, **kwargs):
        try:
            return await coro(*args, **kwargs)
        except RedisError as exc:
            # Any connection and protocol related issues but also invalid
            # Redis-command formatting (not likely).
            log.exception(exc)
        except SocketError:
            log.error("Connection with Redis has been lost. Retrying...")
    return wrapper


class Memory(Service):

    def __init__(self, nyuki):
        self.store = None
        self.config = {}
        self.service = nyuki.config['service']
        self.loop = nyuki.loop or asyncio.get_event_loop()

    @property
    def available(self):
        return self.store is not None

    def key(self, instance, *args):
        keyspace = '{}.{}'.format(self.service, instance)
        for arg in args:
            keyspace = '{}.{}'.format(keyspace, arg)
        return keyspace

    def configure(self, *args, **kwargs):
        self.config = kwargs

    async def start(self, *args, **kwargs):
        """
        Setup a shared memory using Redis.
        """
        self.store = await create_reconnecting_redis(
            (
                self.config.get('host', 'localhost'),
                self.config.get('port', 6379)
            ),
            db=self.config.get('database', 0),
            ssl=self.config.get('ssl'),
            loop=self.loop
        )

        try:
            # An initial 'ping' command allows to immediately check the
            # connection health.
            await self.store.ping()
            log.info("Connection made with Redis")
        except (RedisError, SocketError) as exc:
            log.error("Fail to connect to Redis")
            log.exception(exc)

    async def stop(self, *args, **kwargs):
        pass



import re
import asyncio
import logging
from enum import Enum
from uuid import uuid4
from aiohttp.web import FileField
from tukio import get_broker, EXEC_TOPIC
from tukio.utils import FutureState
from tukio.workflow import WorkflowTemplate, WorkflowExecState
from pymongo import DESCENDING, ASCENDING
from pymongo.errors import AutoReconnect, DuplicateKeyError
from datetime import datetime, timezone
from bson.codec_options import CodecOptions

from nyuki.utils import from_isoformat
from nyuki.api import Response, resource, content_type
from nyuki.workflow.tasks.utils.uri import URI, InvalidWorkflowUri


log = logging.getLogger(__name__)


class Ordering(Enum):

    title_asc = ('title', ASCENDING)
    title_desc = ('title', DESCENDING)
    start_asc = ('exec.start', ASCENDING)
    start_desc = ('exec.start', DESCENDING)
    end_asc = ('exec.end', ASCENDING)
    end_desc = ('exec.end', DESCENDING)

    @classmethod
    def keys(cls):
        return [key for key in cls.__members__.keys()]


class InstanceCollections:

    REQUESTER_REGEX = re.compile(r'^nyuki://.*')
    # TODO: Many fields may not be necessary for the frontend.
    TASK_HISTORY_FILTERS = {
        '_id': 0,
        # Tasks
        'id': 1,
        'name': 1,
        'config': 1,
        'topics': 1,
        'title': 1,
        # Exec
        'exec.id': 1,
        'exec.start': 1,
        'exec.end': 1,
        'exec.state': 1,
        # Graph-specific data fields
        'exec.outputs.quorum': 1,
        'exec.reporting.status': 1,
        'exec.reporting.errors': 1,
    }

    def __init__(self, db):
        # Handle timezones in mongo collections.
        # See http://api.mongodb.com/python/current/examples/datetimes.html#reading-time
        self._workflows = db['workflow-instances'].with_options(
            codec_options=CodecOptions(tz_aware=True, tzinfo=timezone.utc)
        )
        self._tasks = db['task-instances'].with_options(
            codec_options=CodecOptions(tz_aware=True, tzinfo=timezone.utc)
        )
        asyncio.ensure_future(self.index())

    async def index(self):
        # Workflow
        await self._workflows.create_index('exec.id', unique=True)
        await self._workflows.create_index('exec.state')
        await self._workflows.create_index('exec.requester')
        # Search and sorting indexes
        await self._workflows.create_index('title')
        await self._workflows.create_index('exec.start')
        await self._workflows.create_index('exec.end')
        # Task
        await self._tasks.create_index('workflow_exec_id')
        await self._tasks.create_index('exec.id')

    async def _get_wflow_tasks(self, workflow_exec_id, full=False):
        """
        Fetch the list of task execs for this workflow exec.
        """
        if full is False:
            filters = self.TASK_HISTORY_FILTERS
        else:
            filters = {'_id': 0, 'workflow_exec_id': 0}
        cursor = self._tasks.find(
            {'workflow_exec_id': workflow_exec_id},
            filters,
        )
        return await cursor.to_list(None)

    async def get_one(self, exec_id, full=False):
        """
        Return the instance with `exec_id` from workflow history.
        """
        workflow = await self._workflows.find_one(
            {'exec.id': exec_id}, {'_id': 0}
        )
        if workflow:
            workflow['tasks'] = await self._get_wflow_tasks(exec_id, full)
        return workflow

    async def get_task(self, task_id, full=False):
        """
        Return the informations of one single executed task.
        """
        if full is False:
            filters = {'_id': 0, 'exec.inputs': 0, 'exec.outputs': 0}
        else:
            filters = {'_id': 0}
        return await self._tasks.find_one({'exec.id': task_id}, filters)

    async def get_task_data(self, task_id):
        """
        Return the data (inputs/outputs) of one executed task.
        """
        task = await self._tasks.find_one(
            {'exec.id': task_id},
            {'_id': 0, 'exec.inputs': 1, 'exec.outputs': 1},
        )
        return {
            'inputs': task['exec']['inputs'],
            'outputs': task['exec']['outputs'],
        } if task else None

    async def get(self, root=False, full=False, offset=None, limit=None,
                  since=None, state=None, search=None, order=None):
        """
        Return all instances from history from `since` with state `state`.
        """
        query = {}
        # Prepare query
        if isinstance(since, datetime):
            query['exec.start'] = {'$gte': since}
        if isinstance(state, Enum):
            query['exec.state'] = state.value
        if root is True:
            query['exec.requester'] = {'$not': self.REQUESTER_REGEX}
        if search:
            query['title'] = {'$regex': '.*{}.*'.format(search)}

        if full is False:
            fields = {
                '_id': 0,
                'title': 1,
                'exec': 1,
                'id': 1,
                'version': 1,
                'draft': 1
            }
        else:
            fields = {'_id': 0}

        cursor = self._workflows.find(query, fields)
        # Count total results regardless of limit/offset
        count = await cursor.count()

        # Sort depending on Order enum values
        if order is not None:
            cursor.sort(*order)
        else:
            # End descending by default
            cursor.sort(*Ordering.end_desc.value)

        # Set offset and limit
        if isinstance(offset, int) and offset >= 0:
            cursor.skip(offset)
        if isinstance(limit, int) and limit > 0:
            cursor.limit(limit)

        # Execute query
        workflows = await cursor.to_list(None)
        if full is True:
            for workflow in workflows:
                workflow['tasks'] = await self._get_wflow_tasks(
                    workflow['exec']['id'], True
                )
        return count, workflows

    async def insert(self, workflow):
        """
        Insert a finished workflow report into the workflow history.
        """
        # Split tasks exec and workflow exec.
        for task in workflow['tasks']:
            task['workflow_exec_id'] = workflow['exec']['id']
            await self._tasks.insert(task)
        del workflow['tasks']

        try:
            await self._workflows.insert(workflow)
        except DuplicateKeyError:
            # If it's a duplicate, we don't want to lose it
            workflow['exec']['duplicate'] = workflow['exec']['id']
            workflow['exec']['id'] = str(uuid4())
            await self._workflows.insert(workflow)


class _WorkflowResource:

    """
    Share methods between workflow resources
    """

    def register_async_handler(self, async_topic, events, wflow):
        broker = get_broker()
        topic = '/'.join((EXEC_TOPIC, wflow.uid))

        if events:
            events = events.split(',')

        async def exec_handler(event):
            # Publish the event's data if requested.
            if not events or event.data['type'] in events:
                await self.nyuki.bus.publish(event.data, async_topic)
            # If the workflow is in a final state, unregister
            if event.data['type'] in [
                WorkflowExecState.end.value,
                WorkflowExecState.error.value
            ]:
                broker.unregister(exec_handler, topic=topic)

        broker.register(exec_handler, topic=topic)


@resource('/workflow/instances', ['v1'], 'application/json')
class ApiWorkflows(_WorkflowResource):

    async def get(self, request):
        """
        Return workflow instances
        """
        return Response([
            workflow.report()
            for workflow in self.nyuki.running_workflows.values()
        ])

    async def put(self, request):
        """
        Start a workflow from payload:
        {
            "id": "template_id",
            "draft": true/false
            "exec": {}
        }
        """
        async_topic = request.headers.get('X-Surycat-Async-Topic')
        async_events = request.headers.get('X-Surycat-Async-Events')
        exec_track = request.headers.get('X-Surycat-Exec-Track')
        requester = request.headers.get('Referer')
        request = await request.json()

        if 'id' not in request:
            return Response(status=400, body={
                'error': "Template's ID key 'id' is mandatory"
            })
        draft = request.get('draft', False)
        data = request.get('inputs', {})
        exec = request.get('exec')

        if exec:
            # Suspended/crashed instance
            # The request's payload is the last known execution report
            templates = [request]
            if exec['id'] in self.nyuki.running_workflows:
                return Response(status=400, body={
                    'error': 'This workflow is already being rescued'
                })
        else:
            # Fetch the template from the storage
            try:
                templates = await self.nyuki.storage.templates.get(
                    request['id'],
                    draft=draft,
                    with_metadata=True
                )
            except AutoReconnect:
                return Response(status=503)

        if not templates:
            return Response(status=404, body={
                'error': 'Could not find a suitable template to run'
            })

        wf_tmpl = WorkflowTemplate.from_dict(templates[0])
        if exec:
            wflow = await self.nyuki.engine.rescue(wf_tmpl, request)
        elif draft:
            wflow = await self.nyuki.engine.run_once(wf_tmpl, data)
        else:
            wflow = await self.nyuki.engine.trigger(wf_tmpl.uid, data)

        if wflow is None:
            return Response(status=400, body={
                'error': 'Could not start any workflow from this template'
            })

        # Prevent workflow loop
        exec_track = exec_track.split(',') if exec_track else []
        holder = self.nyuki.bus.name
        for ancestor in exec_track:
            try:
                info = URI.parse(ancestor)
            except InvalidWorkflowUri:
                continue
            if info.template_id == wf_tmpl.uid and info.holder == holder:
                return Response(status=400, body={
                    'error': 'Loop detected between workflows'
                })

        # Keep full instance+template in nyuki's memory
        wfinst = self.nyuki.new_workflow(
            templates[0], wflow,
            track=exec_track,
            requester=requester
        )
        # Handle async workflow exec updates
        if async_topic is not None:
            self.register_async_handler(async_topic, async_events, wflow)

        try:
            # Wait up to 30 seconds for the workflow to start.
            await asyncio.wait_for(wfinst.instance._committed.wait(), 30.0)
        except asyncio.TimeoutError:
            status = 201
        else:
            status = 200

        return Response(wfinst.report(), status=status)


@resource('/workflow/instances/{iid}', versions=['v1'])
class ApiWorkflow(_WorkflowResource):

    async def get(self, request, iid):
        """
        Return a workflow instance
        """
        try:
            return Response(self.nyuki.running_workflows[iid].report())
        except KeyError:
            return Response(status=404)

    async def post(self, request, iid):
        """
        Suspend/resume a runnning workflow.
        """
        try:
            wf = self.nyuki.running_workflows[iid]
        except KeyError:
            return Response(status=404)

        request = await request.json()

        try:
            action = request['action']
        except KeyError:
            return Response(status=400, body={
                'action parameter required'
            })

        # Should we return 409 Conflict if the status is already set ?
        if action == 'suspend':
            wf.instance.suspend()
        elif action == 'resume':
            wf.instance.resume()
        else:
            return Response(status=400, body={
                "action must be 'suspend' or 'resume'"
            })

        return Response(wf.report())

    async def delete(self, request, iid):
        """
        Cancel a workflow instance.
        """
        try:
            self.nyuki.running_workflows[iid].instance.cancel()
        except KeyError:
            return Response(status=404)


@resource('/workflow/history', versions=['v1'])
class ApiWorkflowsHistory:

    async def get(self, request):
        """
        Filters:
            * `root` return only the root workflows
            * `full` return the full graph and details of all workflows
                * :warning: can be a huge amount of data
            * `since` return the workflows since this date
            * `state` return the workflows on this FutureState
            * `offset` return the worflows from this offset
            * `limit` return this amount of workflows
            * `order` order results following the Ordering enum values
            * `search` search templates with specific title
        """
        # Filter on start date
        since = request.GET.get('since')
        if since:
            try:
                since = from_isoformat(since)
            except ValueError:
                return Response(status=400, body={
                    'error': "Could not parse date '{}'".format(since)
                })
        # Filter on state value
        state = request.GET.get('state')
        if state:
            try:
                state = FutureState(state)
            except ValueError:
                return Response(status=400, body={
                    'error': "Unknown state '{}'".format(state)
                })
        # Skip first items
        offset = request.GET.get('offset')
        if offset:
            try:
                offset = int(offset)
            except ValueError:
                return Response(status=400, body={
                    'error': 'Offset must be an int'
                })
        # Limit max result
        limit = request.GET.get('limit')
        if limit:
            try:
                limit = int(limit)
            except ValueError:
                return Response(status=400, body={
                    'error': 'Limit must be an int'
                })
        order = request.GET.get('ordering')
        if order:
            try:
                order = Ordering[order].value
            except KeyError:
                return Response(status=400, body={
                    'error': 'Ordering must be in {}'.format(Ordering.keys())
                })

        try:
            count, history = await self.nyuki.storage.instances.get(
                root=(request.GET.get('root') == '1'),
                full=(request.GET.get('full') == '1'),
                search=request.GET.get('search'),
                order=order,
                offset=offset, limit=limit, since=since, state=state,
            )
        except AutoReconnect:
            return Response(status=503)

        data = {'count': count, 'data': history}
        return Response(data)


@resource('/workflow/history/{uid}', versions=['v1'])
class ApiWorkflowHistory:

    async def get(self, request, uid):
        try:
            workflow = await self.nyuki.storage.instances.get_one(
                uid, (request.GET.get('full') == '1')
            )
        except AutoReconnect:
            return Response(status=503)
        if not workflow:
            return Response(status=404)
        return Response(workflow)


@resource('/workflow/history/{uid}/tasks/{task_id}', versions=['v1'])
class ApiWorkflowHistoryTask:

    async def get(self, request, uid, task_id):
        try:
            task = await self.nyuki.storage.instances.get_task(
                task_id, (request.GET.get('full') == '1')
            )
        except AutoReconnect:
            return Response(status=503)
        if not task:
            return Response(status=404)
        return Response(task)


@resource('/workflow/history/{uid}/tasks/{task_id}/data', versions=['v1'])
class ApiWorkflowHistoryTaskData:

    async def get(self, request, uid, task_id):
        try:
            task = await self.nyuki.storage.instances.get_task_data(task_id)
        except AutoReconnect:
            return Response(status=503)
        if not task:
            return Response(status=404)
        return Response(task)


@resource('/workflow/triggers', versions=['v1'])
class ApiWorkflowTriggers:

    async def get(self, request):
        """
        Return the list of all trigger forms
        """
        try:
            triggers = await self.nyuki.storage.triggers.get_all()
        except AutoReconnect:
            return Response(status=503)
        return Response(triggers)

    @content_type('multipart/form-data')
    async def put(self, request):
        """
        Upload a trigger form file
        """
        data = await request.post()
        try:
            form = data['form']
            tid = data['tid']
        except KeyError:
            return Response(status=400, body={
                'error': "'form' and 'tid' are mandatory parameters"
            })
        if not isinstance(form, FileField):
            return Response(status=400, body={
                'error': "'form' field must be a file content"
            })

        content = form.file.read().decode('utf-8')
        try:
            tmpl = await self.nyuki.storage.templates.get(tid)
            if not tmpl:
                return Response(status=404)
            trigger = await self.nyuki.storage.triggers.insert(tid, content)
        except AutoReconnect:
            return Response(status=503)
        return Response(trigger)


@resource('/workflow/triggers/{tid}', versions=['v1'])
class ApiWorkflowTrigger:

    async def get(self, request, tid):
        """
        Return a single trigger form
        """
        try:
            trigger = await self.nyuki.storage.triggers.get(tid)
        except AutoReconnect:
            return Response(status=503)
        if not trigger:
            return Response(status=404)
        return Response(trigger)

    async def delete(self, request, tid):
        """
        Delete a trigger form
        """
        try:
            trigger = await self.nyuki.storage.triggers.get(tid)
        except AutoReconnect:
            return Response(status=503)
        if not trigger:
            return Response(status=404)

        await self.nyuki.storage.triggers.delete(tid)
        return Response(trigger)



import asyncio
import logging
from motor.motor_asyncio import AsyncIOMotorClient

from .api.workflows import InstanceCollections
from .api.templates import TemplateCollection


log = logging.getLogger(__name__)


class _DataProcessingCollection:

    def __init__(self, data_collection):
        self._rules = data_collection
        asyncio.ensure_future(self._rules.create_index('id', unique=True))

    async def get_all(self):
        """
        Return a list of all rules
        """
        cursor = self._rules.find(None, {'_id': 0})
        return await cursor.to_list(None)

    async def get(self, rule_id):
        """
        Return the rule for given id or None
        """
        cursor = self._rules.find({'id': rule_id}, {'_id': 0})
        await cursor.fetch_next
        return cursor.next_object()

    async def insert(self, data):
        """
        Insert a new data processing rule:
        {
            "id": "rule_id",
            "name": "rule_name",
            "config": {
                "some": "configuration"
            }
        }
        """
        query = {'id': data['id']}
        log.info(
            "Inserting data processing rule in collection '%s'",
            self._rules.name
        )
        log.debug('upserting data: %s', data)
        await self._rules.update(query, data, upsert=True)

    async def delete(self, rule_id=None):
        """
        Delete a rule from its id or all rules
        """
        query = {'id': rule_id} if rule_id is not None else None
        log.info("Removing rule(s) from collection '%s'", self._rules.name)
        log.debug('delete query: %s', query)
        await self._rules.remove(query)


class _TriggerCollection:

    def __init__(self, data_collection):
        self._triggers = data_collection
        asyncio.ensure_future(self._triggers.create_index('tid', unique=True))

    async def get_all(self):
        """
        Return a list of all trigger forms
        """
        cursor = self._triggers.find(None, {'_id': 0})
        return await cursor.to_list(None)

    async def get(self, template_id):
        """
        Return the trigger form of a given workflow template id
        """
        cursor = self._triggers.find({'tid': template_id}, {'_id': 0})
        await cursor.fetch_next
        return cursor.next_object()

    async def insert(self, tid, form):
        """
        Insert a trigger form for the given workflow template
        """
        data = {'tid': tid, 'form': form}
        await self._triggers.update({'tid': tid}, data, upsert=True)
        return data

    async def delete(self, template_id=None):
        """
        Delete a trigger form
        """
        query = {'tid': template_id} if template_id is not None else None
        await self._triggers.remove(query)


class MongoStorage:

    DEFAULT_DATABASE = 'workflow'

    def __init__(self, host, database=None, **kwargs):
        log.info("Setting up workflow mongo storage with host '%s'", host)
        client = AsyncIOMotorClient(host, **kwargs)
        db_name = database or self.DEFAULT_DATABASE
        db = client[db_name]
        log.info("Workflow database: '%s'", db_name)

        # Collections
        self.templates = TemplateCollection(db['templates'], db['metadata'])
        self.instances = InstanceCollections(db)
        self.regexes = _DataProcessingCollection(db['regexes'])
        self.lookups = _DataProcessingCollection(db['lookups'])
        self.triggers = _TriggerCollection(db['triggers'])



import json
import asyncio
import logging
import logging.config
from uuid import uuid4
from pijon import Pijon
from jsonschema import validate
from signal import SIGHUP, SIGINT, SIGTERM

from .api import Api
from .api.bus import ApiBusReplay, ApiBusTopics, ApiBusPublish
from .api.config import ApiConfiguration, ApiSwagger
from .bus import XmppBus, MqttBus, reporting
from .commands import get_command_kwargs
from .config import get_full_config, write_conf_json, merge_configs
from .debugging import StackSampler, ApiSampleEmitter
from .logs import DEFAULT_LOGGING
from .services import ServiceManager
from .websocket import WebsocketHandler
from .discovery import Discovery
from .raft import RaftProtocol, ApiRaft
from .memory import Memory


log = logging.getLogger(__name__)


class Nyuki:

    """
    A lightweigh base class to build nyukis. A nyuki provides tools that shall
    help the developer with managing the following topics:
      - Bus of communication between nyukis.
      - Asynchronous events.
      - Capabilities exposure through a REST API.
    This class has been written to perform the features above in a reliable,
    single-threaded, asynchronous and concurrent-safe environment.
    The core engine of a nyuki implementation is the asyncio event loop
    (a single loop is used for all features).
    """

    # Configuration schema must follow jsonschema rules.
    BASE_CONF_SCHEMA = {
        'type': 'object',
        'required': ['log'],
        'properties': {
            'service': {'type': 'string', 'minLength': 1},
            'trace': {'type': 'boolean'},
        }
    }
    # API endpoints
    HTTP_RESOURCES = [
        ApiBusPublish,
        ApiBusReplay,
        ApiBusTopics,
        ApiConfiguration,
        ApiSwagger,
        ApiRaft,
        ApiSampleEmitter,
    ]

    def __init__(self, **kwargs):
        # List of configuration schemas
        self._schemas = []
        self._id = str(uuid4())[:8]

        # Initialize logging
        logging.config.dictConfig(DEFAULT_LOGGING)

        # Get configuration from multiple sources and register base schema
        kwargs = kwargs or get_command_kwargs()
        # Storing the optional init params, will be used when reloading
        self._launch_params = kwargs
        self._config_filename = kwargs.get('config')
        self._config = get_full_config(**kwargs)
        if self._config['log'] != DEFAULT_LOGGING:
            logging.config.dictConfig({
                **DEFAULT_LOGGING,
                **self._config['log']
            })
        self.register_schema(self.BASE_CONF_SCHEMA)

        # Setup stack sampling
        self._sampler = None
        self._set_stack_sampling()
        # Set loop
        self.loop = asyncio.get_event_loop() or asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

        self._services = ServiceManager(self)
        self._services.add('api', Api(self))

        # Add bus service if in conf file, xmpp (default) or mqtt
        bus_config = self._config.get('bus')
        if bus_config is not None:
            bus_service = bus_config.get('service', 'xmpp')
            if bus_service == 'xmpp':
                self._services.add('bus', XmppBus(self))
            elif bus_service == 'mqtt':
                self._services.add('bus', MqttBus(self))
        # Add websocket server if in conf file
        if self._config.get('websocket') is not None:
            self._services.add('websocket', WebsocketHandler(self))

        # Add NaaS (nyuki-as-a-service) related services
        if self._config.get('service'):
            # Discovery
            method = self._config.get('discovery', {}).get('method', 'dns')
            discovery = Discovery.get(method)
            self._services.add('discovery', discovery(self))
            # Raft
            self._services.add('raft', RaftProtocol(self))
            self.discovery.register(self.raft.discovery_handler)
            # Memory
            if self._config.get('memory'):
                self._services.add('memory', Memory(self))

        self.is_stopping = False

    def __getattribute__(self, name):
        """
        Getattr, or returns the specified service
        """
        try:
            return super().__getattribute__(name)
        except AttributeError as exc:
            try:
                return self._services.get(name)
            except KeyError:
                raise exc

    @property
    def id(self):
        return self._id

    @property
    def config(self):
        return self._config

    @property
    def reporter(self):
        """
        Ensure backwards compatibility
        TODO: must be removed
        """
        log.warning('Deprecated reporting call, use nyuki.bus.reporting')
        return reporting

    def _exception_handler(self, loop, context):
        if 'exception' not in context:
            log.warning('No exception in context: %s', context)
            return
        log.debug('Exception context: %s', context)

        exc = Exception("could not retrieve exception's traceback")
        if 'future' in context:
            try:
                context['future'].result()
            except Exception as e:
                exc = e
        else:
            exc = context['exception']

        reporting.exception(exc)

    def start(self):
        """
        Start the nyuki
        The nyuki process is terminated when this method is finished
        """
        self.loop.add_signal_handler(SIGTERM, self.abort, SIGTERM)
        self.loop.add_signal_handler(SIGINT, self.abort, SIGINT)
        self.loop.add_signal_handler(SIGHUP, self.hang_up, SIGHUP)

        # Configure services with nyuki's configuration
        log.debug('Running configure for services')
        for name, service in self._services.all.items():
            service.configure(**self._config.get(name, {}))
        log.debug('Done configuring')

        # Start services
        self.loop.run_until_complete(self._services.start())

        if 'bus' in self._services:
            self.bus.init_reporting()
            self.loop.set_exception_handler(self._exception_handler)

        # Call for setup
        if not asyncio.iscoroutinefunction(self.setup):
            log.warning('setup method must be a coroutine')
            self.setup = asyncio.coroutine(self.setup)
        self.loop.run_until_complete(self.setup())

        # Main loop
        self.loop.run_forever()

        # Call for teardown
        if not asyncio.iscoroutinefunction(self.teardown):
            log.warning('teardown method must be a coroutine')
            self.teardown = asyncio.coroutine(self.teardown)
        self.loop.run_until_complete(self.teardown())

        # Close everything : terminates nyuki
        self.loop.close()

    def _stop_loop(self):
        """
        Call the loop to stop itself.
        """
        self.loop.call_soon_threadsafe(self.loop.stop)

    def abort(self, signal):
        """
        Signal handler: gracefully stop the nyuki.
        """
        log.warning('Caught signal %d, stopping nyuki', signal)
        asyncio.ensure_future(self.stop())

    async def stop(self, timeout=5):
        """
        Stop the nyuki
        """
        if self.is_stopping:
            log.warning('Force closing the nyuki')
            self._stop_loop()
            return

        self.is_stopping = True
        if self._sampler:
            self._sampler.stop()
        await self._services.stop()
        self._stop_loop()

    def hang_up(self, signal):
        """
        Signal handler: reload the nyuki
        """
        log.warning('Caught signal %d, reloading the nyuki', signal)
        try:
            self._config = get_full_config(**self._launch_params)
        except json.decoder.JSONDecodeError as e:
            log.error(
                'Could not load the new configuration, '
                'fallback on the previous one: "%s"', e
            )
        else:
            asyncio.ensure_future(self._reload_config())

    def register_schema(self, schema, format_checker=None):
        """
        Add a jsonschema to validate on configuration update.
        """
        self._schemas.append((schema, format_checker))
        self._validate_config()

    def _validate_config(self, config=None):
        """
        Validate on all registered configuration schemas.
        """
        log.debug('Validating configuration')
        config = config or self._config
        for schema, checker in self._schemas:
            validate(config, schema, format_checker=checker)

    async def on_buffer_full(self, free_slot):
        """
        Called when the bus memory buffer is full of published events.
        """
        asyncio.ensure_future(self.buffer_full())
        await free_slot.wait()
        asyncio.ensure_future(self.free_slot())

    async def setup(self):
        """
        First thing called when starting the event loop, coroutine or not.
        """
        log.warning('Setup called, but not overridden')

    async def reload(self):
        """
        Called when the configuration is modified
        """
        log.warning('Reload called, but not overridden')

    async def teardown(self):
        """
        Called right before closing the event loop, stopping the Nyuki.
        """
        log.warning('Teardown called, but not overridden')

    async def buffer_full(self):
        """
        Called when the bus memory buffer is full.
        """
        log.warning('Buffer full callback not overridden')

    async def free_slot(self):
        """
        Called when the bus memory buffer has free slots available.
        """
        log.warning('Buffer free slot callback not overridden')

    def update_config(self, *new_confs):
        """
        Update the current configuration with the given list of dicts.
        """
        config = merge_configs(self._config, *new_confs)
        self._validate_config(config)
        self._config = config

    def save_config(self):
        """
        Save the current configuration dict to its JSON file.
        """
        if self._config_filename:
            write_conf_json(self.config, self._config_filename)
        else:
            log.warning('Not saving the default read-only configuration file')

    def migrate_config(self):
        """
        Migrate configuration dict using pijon migrations.
        """
        tool = Pijon(load=False)
        current = self._config.get('version', 0)
        log.debug("Current configuration file version: {}".format(current))

        if not tool.migrations or current >= tool.last_migration:
            log.debug("Configuration is up to date")
            return

        log.warning("Configuration seems out of date, applying migrations")
        update = tool.migrate(self._config)
        self._validate_config(update)
        self._config = update
        self.save_config()

    async def _reload_config(self, request=None):
        """
        Reload the configuration and the services
        """
        logging.config.dictConfig(self._config['log'])
        self._set_stack_sampling()
        await self.reload()
        for name, service in self._services.all.items():
            if (request is not None and name in request) or request is None:
                await service.stop()
                service.configure(**self._config.get(name, {}))
                asyncio.ensure_future(service.start())

    def _set_stack_sampling(self):
        enable = self.config.get('trace') is True
        # Enable sampler trace
        if not self._sampler and enable:
            self._sampler = StackSampler()
            self._sampler.start()
        # Disable sampler trace if it is enabled
        elif self._sampler and not enable:
            self._sampler.stop()
            self._sampler = None



import asyncio
import json
import logging
import websockets
from jsonschema import validate, ValidationError

from nyuki.services import Service
from nyuki.utils import serialize_object


log = logging.getLogger(__name__)


class WebsocketResource:

    # Value before the client is timed out after not sending any keepalive.
    KEEPALIVE = 60

    def __init__(self, path):
        self._clients = []
        self._path = path
        log.info("New websocket resource on path: '%s'", path)
        WebsocketHandler.RESOURCES[path] = self

    def end(self):
        """
        Unregister this resource, the object becoming stall.
        """
        asyncio.ensure_future(self.close_clients())
        log.info("Ended websocket resource on path: '%s'", self._path)
        del WebsocketHandler.RESOURCES[self._path]

    async def close_clients(self):
        log.debug("Closing all websocket connections on path: '%s'", self._path)
        tasks = [
            asyncio.ensure_future(self.remove_client(client, 1001, 'server closing'))
            for client in self._clients
        ]
        if tasks:
            await asyncio.wait(tasks)
        self._clients = []

    async def add_client(self, client):
        ready = {
            'type': 'ready',
            'keepalive_delay': self.KEEPALIVE,
            'data': await self.ready(client) or {}
        }
        await client.send(json.dumps(ready, default=serialize_object))
        self._clients.append(client)

    async def remove_client(self, client, code=None, reason=None):
        if client not in self._clients:
            return
        if code is None:
            code = 1000
        if reason is None:
            reason = 'connection closed normally'
        await self.close(client)
        await client.close(code, reason)
        if client in self._clients:
            self._clients.remove(client)

    async def broadcast(self, data, timeout=None):
        """
        Send a message to every connected client.
        """
        if not self._clients:
            return
        if not isinstance(data, str):
            data = json.dumps(data, default=serialize_object)

        tasks = [
            asyncio.ensure_future(client.send(data))
            for client in self._clients
        ]
        log.debug(
            'Sending data of length %s to %s clients',
            len(data), len(self._clients)
        )
        await asyncio.wait(tasks, timeout=timeout)

    async def ready(self, client):
        """
        Called on a new client connection.
        """

    async def close(self, client):
        """
        Called on a client disconnection.
        """


class WebsocketHandler(Service):

    CONF_SCHEMA = {
        'type': 'object',
        'properties': {
            'websocket': {
                'type': 'object',
                'properties': {
                    'host': {'type': 'string'},
                    'port': {'type': 'integer', 'minimum': 1}
                }
            }
        }
    }

    RESOURCES = {}
    KEEPALIVE_SCHEMA = {
        'type': 'object',
        'required': ['type'],
        'properties': {
            'type': {
                'type': 'string',
                'minLength': 1
            }
        }
    }

    def __init__(self, nyuki, loop=None):
        self._nyuki = nyuki
        self._nyuki.register_schema(self.CONF_SCHEMA)
        self._loop = loop or asyncio.get_event_loop()
        self.host = None
        self.port = None
        self.server = None

    async def start(self):
        """
        Start the websocket server
        """
        log.info("Starting websocket server on %s:%s", self.host, self.port)
        self.server = await websockets.serve(
            self._wshandler, self.host, self.port
        )

    def configure(self, host='0.0.0.0', port=5559):
        self.host = host
        self.port = port

    async def stop(self):
        if not self.server:
            return
        for resource in self.RESOURCES.values():
            asyncio.ensure_future(resource.close_clients())
        self.server.close()
        await self.server.wait_closed()

    def _schedule_keepalive(self, resource, client):
        def timeout():
            asyncio.ensure_future(resource.remove_client(
                client, 4008, 'keepalive timed out'
            ))
        return self._loop.call_later(resource.KEEPALIVE, timeout)

    async def _wshandler(self, websocket, path):
        """
        Main handler for a newly connected client
        """
        try:
            resource = self.RESOURCES[path]
        except KeyError:
            await websocket.close(4004, 'resource not found')
            return

        await resource.add_client(websocket)
        handle = self._schedule_keepalive(resource, websocket)

        while True:
            # Main read loop
            try:
                message = await websocket.recv()
            except websockets.exceptions.ConnectionClosed as exc:
                log.debug('client connection closed: %s', exc)
                break

            # Decode JSON message
            try:
                data = json.loads(message)
            except ValueError:
                log.debug('Message received not JSON: %s', message)
                continue

            try:
                validate(data, self.KEEPALIVE_SCHEMA)
            except ValidationError:
                log.debug('Invalid keepalive received: %s', data)
                continue

            mtype = data['type']
            if mtype == 'keepalive':
                handle.cancel()
                handle = self._schedule_keepalive(resource, websocket)

        # 'websocket' client may already be closed here.
        handle.cancel()
        await resource.remove_client(websocket)



from nyuki.services import Service


class Discovery(type):

    _REGISTRY = {}

    def __new__(mcs, name, bases, attrs):
        new_mcs = type.__new__(mcs, name, bases, attrs)
        if new_mcs.SCHEME:
            mcs._REGISTRY[new_mcs.SCHEME] = new_mcs
        return new_mcs

    @classmethod
    def get(mcs, name):
        method = mcs._REGISTRY.get(name)
        if not method:
            raise ValueError("Unknown discovery method '{}'".format(name))
        return mcs._REGISTRY[name]


class DiscoveryService(Service, metaclass=Discovery):

    SERVICE = 'discovery'
    SCHEME = None

    def register(self, callback):
        raise NotImplementedError()


from .dns import DnsDiscovery



import os
import sys
import socket
import asyncio
import logging
from traceback import TracebackException

from nyuki.utils import from_isoformat, utcnow


log = logging.getLogger(__name__)


class Reporter(object):

    """
    Mqtt and Xmpp bus are handled differently as the topics mechanism
    is not the same.
    """

    EXCEPTION_TTL = 3600
    MONIT_TOPIC = '+/monitoring'

    def __init__(self):
        self._name = None
        self._loop = None
        self._publisher = None
        self._channel = None
        self._handler = None
        self._last_exceptions = list()

    def init(self, name, publisher, loop=None):
        self._name = name
        self._loop = loop or asyncio.get_event_loop()
        self._publisher = publisher
        self._service = self._publisher.SERVICE

        if self._service == 'xmpp':
            self._channel = 'monitoring'
        elif self._service == 'mqtt':
            self._channel = self.MONIT_TOPIC.replace('+', self._name)
        else:
            raise TypeError('Nyuki publisher must be XmppBus or MqttBus')

    async def _handle_report(self, topic, data):
        """
        Handle XMPP report, ignore if it comes from this reporter
        """
        if self._handler is None:
            log.debug('Report received, no handler set')
            return

        if data['author'] == self._name:
            log.debug('Received own report, ignoring')
            return

        await self._handler(topic, data)

    def register_handler(self, handler):
        """
        Register all required handlers for received reports
        """
        if not asyncio.iscoroutinefunction(handler):
            raise ValueError('handler must be a coroutine')
        if self._service == 'mqtt':
            asyncio.ensure_future(self._publisher.subscribe(
                self.MONIT_TOPIC, self._handle_report
            ))
        self._handler = handler

    def send_report(self, rtype, data):
        """
        Send reports with a type and any data

        Using docker/fleet, we require some informations about IP/hostname of
        our nyuki containers, using environnement vars :
            - MACHINE_NAME: machine hostname
            - DEFAULT_IPV4: local container ipv4
        Otherwise, the nyuki try and search for it by itself.
        """
        if not self._publisher:
            log.warning('Reporting not initiated')
            return

        report = {
            'hostname': os.environ.get('MACHINE_NAME', socket.gethostname()),
            'ipv4': os.environ.get(
                'DEFAULT_IPV4',
                socket.gethostbyname(socket.gethostname())
            ),
            'type': rtype,
            'author': self._name,
            'datetime': utcnow(),
            'data': data
        }
        log.info("Sending report data with type '%s'", rtype)
        asyncio.ensure_future(self._publisher.publish(report, self._channel))

    def exception(self, exc):
        """
        Helper to report an exception traceback from its object
        """
        traceback = TracebackException.from_exception(exc)
        formatted = ''.join(traceback.format())
        log.error(formatted)

        if formatted in self._last_exceptions:
            log.debug('Exception already logged')
            return

        # Retain the formatted exception in memory to avoid looping
        self._last_exceptions.append(formatted)
        self._loop.call_later(
            self.EXCEPTION_TTL, self._forget_exception, formatted
        )

        self.send_report('exception', {'traceback': formatted})

    def _forget_exception(self, formatted):
        self._last_exceptions.remove(formatted)


sys.modules[__name__] = Reporter()



import asyncio
import logging
from pymongo import DESCENDING
from pymongo.errors import AutoReconnect, DuplicateKeyError
from tukio.workflow import TemplateGraphError, WorkflowTemplate

from nyuki.api import Response, resource
from nyuki.workflow.validation import validate, TemplateError


log = logging.getLogger(__name__)


class ConflictError(Exception):
    pass


class DuplicateTemplateError(Exception):
    pass


class TemplateCollection:

    """
    Holds all the templates created for tukio, with their versions.
    These records will be used to ensure a persistence of created workflows
    in case the nyuki get into trouble.
    Templates are retrieved and loaded at startup.
    """

    def __init__(self, templates_collection, metadata_collection):
        self._templates = templates_collection
        self._metadata = metadata_collection
        # Indexes (ASCENDING by default)
        asyncio.ensure_future(self._metadata.create_index('id', unique=True))
        asyncio.ensure_future(self._templates.create_index(
            [('id', DESCENDING), ('version', DESCENDING)],
            unique=True
        ))
        asyncio.ensure_future(self._templates.create_index(
            [('id', DESCENDING), ('draft', DESCENDING)]
        ))

    async def get_metadata(self, tid=None):
        """
        Return metadata
        """
        query = {'id': tid} if tid else None
        cursor = self._metadata.find(query, {'_id': 0})
        return await cursor.to_list(None)

    async def get_all(self, full=False, latest=False, draft=False, with_metadata=True):
        """
        Return all templates, used at nyuki's startup and GET /v1/templates
        Fetch latest versions if latest=True
        Fetch drafts if draft=True
        Both drafts and latest version if both are True
        """
        filters = {'_id': 0}
        # '/v1/workflow/templates' does not requires all the informations
        if full is False:
            filters.update({'id': 1, 'draft': 1, 'version': 1, 'topics': 1})

        cursor = self._templates.find(None, filters)
        cursor.sort('version', DESCENDING)
        templates = await cursor.to_list(None)

        # Collect metadata
        if with_metadata and templates:
            metadatas = await self.get_metadata()
            metadatas = {meta['id']: meta for meta in metadatas}
            for template in templates:
                template.update(metadatas[template['id']])

        if latest is False and draft is False:
            return templates

        # Retrieve the latest versions + drafts
        lasts = {}
        drafts = []

        for template in templates:
            if draft and template['draft']:
                drafts.append(template)
            elif latest and not template['draft'] and template['id'] not in lasts:
                lasts[template['id']] = template

        return drafts + list(lasts.values())

    async def get(self, tid, version=None, draft=None, with_metadata=True):
        """
        Return a template's configuration and versions
        """
        query = {'id': tid}

        if version:
            query['version'] = int(version)
        if draft is not None:
            query['draft'] = draft

        cursor = self._templates.find(query, {'_id': 0})
        cursor.sort('version', DESCENDING)
        templates = await cursor.to_list(None)

        # Collect metadata
        if with_metadata and templates:
            metadatas = await self.get_metadata(tid)
            if metadatas:
                for template in templates:
                    template.update(metadatas[0])

        return templates

    async def get_last_version(self, tid):
        """
        Return the highest version of a template
        """
        query = {'id': tid, 'draft': False}
        cursor = self._templates.find(query)
        cursor.sort('version', DESCENDING)
        await cursor.fetch_next

        template = cursor.next_object()
        return template['version'] if template else 0

    async def insert(self, template):
        """
        Insert a template dict, not updatable
        """
        query = {
            'id': template['id'],
            'version': template['version']
        }

        # Remove draft if any
        await self.delete(template['id'], template['version'], True)

        log.info('Insert template with query: %s', query)
        try:
            # Copy dict, mongo somehow alter the given dict
            await self._templates.insert(template.copy())
        except DuplicateKeyError as exc:
            raise DuplicateTemplateError from exc

    async def insert_draft(self, template):
        """
        Check and insert draft, updatable
        """
        query = {
            'id': template['id'],
            'draft': True
        }

        try:
            log.info('Update draft for query: %s', query)
            await self._templates.update(query, template, upsert=True)
        except DuplicateKeyError as exc:
            raise DuplicateTemplateError from exc

    async def insert_metadata(self, metadata):
        """
        Check and insert metadata
        """
        query = {'id': metadata['id']}

        metadata = {
            'id': metadata['id'],
            'title': metadata.get('title', ''),
            'tags': metadata.get('tags', [])
        }

        log.info('Update metadata for query: %s', query)
        await self._metadata.update(query, metadata, upsert=True)

        return metadata

    async def publish_draft(self, tid):
        """
        From draft to production
        """
        query = {'id': tid, 'draft': True}
        await self._templates.update(query, {'$set': {'draft': False}})

    async def delete(self, tid, version=None, draft=None):
        """
        Delete a template from its id with all its versions
        """
        query = {'id': tid}
        if version:
            query['version'] = version
        if draft is not None:
            query['draft'] = draft

        log.info("Removing template(s) with query: %s", query)

        await self._templates.remove(query)
        left = await self._templates.find({'id': tid}).count()
        if not left:
            await self._metadata.remove({'id': tid})


@resource('/workflow/tasks', versions=['v1'])
class ApiTasks:

    async def get(self, request):
        """
        Return the available tasks
        """
        return Response(self.nyuki.AVAILABLE_TASKS)


class _TemplateResource:

    """
    Share methods between templates resources
    """

    async def update_draft(self, template, from_request=None):
        """
        Helper to insert/update a draft
        """
        tmpl_dict = template.as_dict()

        # Auto-increment version, draft only
        last_version = await self.nyuki.storage.templates.get_last_version(
            template.uid
        )
        tmpl_dict['version'] = last_version + 1
        tmpl_dict['draft'] = True

        # Store task extra info (ie. title)
        if from_request is not None:
            rqst_tasks = from_request.get('tasks', [])
            tmpl_tasks = tmpl_dict['tasks']
            for src in rqst_tasks:
                match = list(filter(lambda t: t['id'] == src['id'], tmpl_tasks))
                if match:
                    match[0].update({'title': src.get('title')})

        try:
            await self.nyuki.storage.templates.insert_draft(tmpl_dict)
        except DuplicateTemplateError as exc:
            raise ConflictError('Template already exists for this version') from exc

        return tmpl_dict

    def errors_from_validation(self, template):
        """
        Validate and return the list of errors if any
        """
        errors = None
        try:
            validate(template)
        except TemplateError as err:
            errors = err.as_dict()
        return errors


@resource('/workflow/templates', versions=['v1'])
class ApiTemplates(_TemplateResource):

    async def get(self, request):
        """
        Return available workflows' DAGs
        """
        try:
            templates = await self.nyuki.storage.templates.get_all(
                full=(request.GET.get('full') == '1'),
                latest=(request.GET.get('latest') == '1'),
                draft=(request.GET.get('draft') == '1'),
            )
        except AutoReconnect:
            return Response(status=503)
        return Response(templates)

    async def put(self, request):
        """
        Create a workflow DAG from JSON
        """
        request = await request.json()

        if 'id' in request:
            try:
                draft = await self.nyuki.storage.templates.get(
                    request['id'], draft=True
                )
            except AutoReconnect:
                return Response(status=503)
            if draft:
                return Response(status=409, body={
                    'error': 'draft already exists'
                })

        if self.nyuki.DEFAULT_POLICY is not None and 'policy' not in request:
            request['policy'] = self.nyuki.DEFAULT_POLICY

        try:
            template = WorkflowTemplate.from_dict(request)
        except TemplateGraphError as exc:
            return Response(status=400, body={
                'error': str(exc)
            })

        try:
            metadata = await self.nyuki.storage.templates.get_metadata(template.uid)
        except AutoReconnect:
            return Response(status=503)
        if not metadata:
            if 'title' not in request:
                return Response(status=400, body={
                    'error': "workflow 'title' key is mandatory"
                })

            metadata = {
                'id': template.uid,
                'title': request['title'],
                'tags': request.get('tags', [])
            }

            await self.nyuki.storage.templates.insert_metadata(metadata)
        else:
            metadata = metadata[0]

        try:
            tmpl_dict = await self.update_draft(template, request)
        except ConflictError as exc:
            return Response(status=409, body={
                'error': exc
            })

        return Response({
            **tmpl_dict,
            **metadata,
            'errors': self.errors_from_validation(template)
        })


@resource('/workflow/templates/{tid}', versions=['v1'])
class ApiTemplate(_TemplateResource):

    async def get(self, request, tid):
        """
        Return the latest version of the template
        """
        try:
            tmpl = await self.nyuki.storage.templates.get(tid)
        except AutoReconnect:
            return Response(status=503)
        if not tmpl:
            return Response(status=404)

        return Response(tmpl)

    async def put(self, request, tid):
        """
        Create a new draft for this template id
        """
        try:
            versions = await self.nyuki.storage.templates.get(tid)
        except AutoReconnect:
            return Response(status=503)
        if not versions:
            return Response(status=404)

        for v in versions:
            if v['draft'] is True:
                return Response(status=409, body={
                    'error': 'This draft already exists'
                })

        request = await request.json()

        try:
            # Set template ID from url
            template = WorkflowTemplate.from_dict({**request, 'id': tid})
        except TemplateGraphError as exc:
            return Response(status=400, body={
                'error': str(exc)
            })

        try:
            tmpl_dict = await self.update_draft(template, request)
        except ConflictError as exc:
            return Response(status=409, body={
                'error': exc
            })

        metadata = await self.nyuki.storage.templates.get_metadata(template.uid)
        metadata = metadata[0]

        return Response({
            **tmpl_dict,
            **metadata,
            'errors': self.errors_from_validation(template)
        })

    async def patch(self, request, tid):
        """
        Modify the template's metadata
        """
        try:
            tmpl = await self.nyuki.storage.templates.get(tid)
        except AutoReconnect:
            return Response(status=503)
        if not tmpl:
            return Response(status=404)

        request = await request.json()

        # Add ID, request dict cleaned in storage
        metadata = await self.nyuki.storage.templates.insert_metadata({
            **request,
            'id': tid
        })

        return Response(metadata)

    async def delete(self, request, tid):
        """
        Delete the template
        """
        try:
            tmpl = await self.nyuki.storage.templates.get(tid)
        except AutoReconnect:
            return Response(status=503)
        if not tmpl:
            return Response(status=404)

        await self.nyuki.storage.templates.delete(tid)
        await self.nyuki.storage.triggers.delete(tid)

        try:
            await self.nyuki.engine.unload(tid)
        except KeyError as exc:
            log.debug(exc)

        return Response(tmpl)


@resource('/workflow/templates/{tid}/{version:\d+}', versions=['v1'])
class ApiTemplateVersion(_TemplateResource):

    async def get(self, request, tid, version):
        """
        Return the template's given version
        """
        try:
            tmpl = await self.nyuki.storage.templates.get(tid, version, False)
        except AutoReconnect:
            return Response(status=503)
        if not tmpl:
            return Response(status=404)

        return Response(tmpl)

    async def delete(self, request, tid, version):
        """
        Delete a template with given version
        """
        try:
            tmpl = await self.nyuki.storage.templates.get(tid)
        except AutoReconnect:
            return Response(status=503)
        if not tmpl:
            return Response(status=404)

        await self.nyuki.storage.templates.delete(tid, version)
        return Response(tmpl[0])


@resource('/workflow/templates/{tid}/draft', versions=['v1'])
class ApiTemplateDraft(_TemplateResource):

    async def get(self, request, tid):
        """
        Return the template's draft, if any
        """
        try:
            tmpl = await self.nyuki.storage.templates.get(tid, draft=True)
        except AutoReconnect:
            return Response(status=503)
        if not tmpl:
            return Response(status=404)

        return Response(tmpl[0])

    async def post(self, request, tid):
        """
        Publish a draft into production
        """
        try:
            tmpl = await self.nyuki.storage.templates.get(tid, draft=True)
        except AutoReconnect:
            return Response(status=503)
        if not tmpl:
            return Response(status=404)

        draft = {
            **tmpl[0],
            'draft': False
        }

        try:
            # Set template ID from url
            template = WorkflowTemplate.from_dict({**draft, 'draft': False})
        except TemplateGraphError as exc:
            return Response(status=400, body={
                'error': str(exc)
            })

        errors = self.errors_from_validation(template)
        if errors is not None:
            return Response(status=400, body=errors)

        await self.nyuki.engine.load(template)
        # Update draft into a new template
        await self.nyuki.storage.templates.publish_draft(tid)
        return Response(draft)

    async def patch(self, request, tid):
        """
        Modify the template's draft
        """
        try:
            tmpl = await self.nyuki.storage.templates.get(tid, draft=True)
        except AutoReconnect:
            return Response(status=503)
        if not tmpl:
            return Response(status=404)

        request = await request.json()

        try:
            # Set template ID from url
            template = WorkflowTemplate.from_dict({**request, 'id': tid})
        except TemplateGraphError as exc:
            return Response(status=400, body={
                'error': str(exc)
            })

        try:
            tmpl_dict = await self.update_draft(template, request)
        except ConflictError as exc:
            return Response(status=409, body={
                'error': str(exc)
            })

        metadata = await self.nyuki.storage.templates.get_metadata(template.uid)
        metadata = metadata[0]

        return Response({
            **tmpl_dict,
            **metadata,
            'errors': self.errors_from_validation(template)
        })

    async def delete(self, request, tid):
        """
        Delete the template's draft
        """
        try:
            tmpl = await self.nyuki.storage.templates.get(tid, draft=True)
        except AutoReconnect:
            return Response(status=503)
        if not tmpl:
            return Response(status=404)

        await self.nyuki.storage.templates.delete(tid, draft=True)
        return Response(tmpl[0])



from aiohttp import web
from aiohttp.hdrs import METH_ALL
import asyncio
from functools import partial
import json
import logging

from nyuki.bus import reporting
from nyuki.services import Service
from nyuki.utils import serialize_object


log = logging.getLogger(__name__)
# aiohttp needs its own logger, and always prints HTTP hits using INFO level
access_log = logging.getLogger('.'.join([__name__, 'access']))
access_log.info = access_log.debug


def resource(path, versions=None, content_type='application/json'):
    """
    Nyuki resource decorator to register a route.
    A resource has multiple HTTP methods (get, post, etc).
    """
    def decorated(cls):
        cls.RESOURCE_CLASS = ResourceClass(cls, path, versions, content_type)
        return cls
    return decorated


def content_type(content_type):
    """
    Decorator to set a specific content_type to one method (get, post...)
    """
    def decorated(func):
        func.CONTENT_TYPE = content_type
        return func
    return decorated


class HTTPBreak(Exception):

    def __init__(self, status, body=None):
        self.status = status
        self.body = body


class Response(web.Response):

    """
    Overrides aiohttp's response to facilitate its usage
    """

    ENCODING = 'utf-8'

    def __init__(self, body=None, **kwargs):

        # Check json
        if isinstance(body, dict) or isinstance(body, list):
            body = json.dumps(body, default=serialize_object).encode(self.ENCODING)
            if not self._get_content_type(kwargs):
                kwargs['content_type'] = 'application/json'
        # Check body
        elif body is not None:
            body = str(body).encode(self.ENCODING)
            if not self._get_content_type(kwargs):
                kwargs['content_type'] = 'text/plain'

        return super().__init__(body=body, **kwargs)

    def _get_content_type(self, kwargs):
        return kwargs.get('content_type') or kwargs.get('headers', {}).get('Content-Type')


async def mw_capability(app, capa_handler):
    """
    Transform the request data to be passed through a capability and
    convert the result into a web response.
    """
    POST_METHODS = web.Request.POST_METHODS - {'DELETE'}

    async def middleware(request):
        # Ensure a content-type check is necessary
        # aiohttp includes DELETE in post methods, we don't want that
        if request.method in POST_METHODS and getattr(capa_handler, 'CONTENT_TYPE', None):
            ctype = capa_handler.CONTENT_TYPE
            # Check content_type from @resource or @content_type decorators
            request_types = request.headers.get('Content-Type', '').split(';')
            required_types = ctype.split(';')
            for required in required_types:
                if required not in request_types:
                    log.debug(
                        "content-type '%s' required. Received '%s'",
                        required, request_types
                    )
                    return Response({'error': 'Wrong or Missing content-type'}, status=400)

            # Check application/json is really a JSON body
            if 'application/json' in required_types:
                try:
                    await request.json()
                except json.decoder.JSONDecodeError:
                    log.debug('request body for application/json must be JSON')
                    return Response(
                        {'error': 'application/json requires a JSON body'},
                        status=400
                    )

            # Check multipart/form-data is really a post form
            if 'multipart/form-data' in required_types:
                try:
                    await request.post()
                except ValueError as exc:
                    log.debug(exc)
                    return Response(status=400, body={
                        'error': 'multipart/form-data must be a form'
                    })

        try:
            capa_resp = await capa_handler(request, **request.match_info)
        except (web.HTTPNotFound, web.HTTPMethodNotAllowed):
            # Avoid sending a report on a simple 404/405
            raise
        except HTTPBreak as exc:
            return Response(exc.body, status=exc.status)
        except Exception as exc:
            reporting.exception(exc)
            raise

        if capa_resp and isinstance(capa_resp, Response):
            return capa_resp
        return Response()

    return middleware


class ResourceClass:

    """
    Allow the extensivity of the nyuki's HTTP resources using the webserver's
    router from the `Api` class.
    """

    def __init__(self, cls, path, versions, content_type):
        self.cls = cls
        self.path = path
        self.versions = versions
        self.content_type = content_type

    def _add_routes(self, router, path):
        resource = router.add_resource(path)
        cls_instance = self.cls()
        for method in METH_ALL:
            handler = getattr(self.cls, method.lower(), None)
            if handler is not None:
                async_handler = asyncio.coroutine(partial(handler, cls_instance))
                async_handler.CONTENT_TYPE = getattr(
                    handler, 'CONTENT_TYPE', self.content_type
                )
                route = resource.add_route(method, async_handler)
                log.debug('Added route: %s', route)

    def register(self, nyuki, router):
        self.cls.nyuki = nyuki
        if not self.versions:
            self._add_routes(router, self.path)
        else:
            for version in self.versions:
                route = '/{}{}'.format(version, self.path)
                self._add_routes(router, route)


class Api(Service):

    """
    Manage a webserver built using the nyuki's defined HTTP resources
    """

    CONF_SCHEMA = {
        "type": "object",
        "required": ["api"],
        "properties": {
            "api": {
                "type": "object",
                "properties": {
                    "host": {"type": "string"},
                    "port": {"type": "integer"}
                }
            }
        }
    }

    def __init__(self, nyuki):
        self._nyuki = nyuki
        self._nyuki.register_schema(self.CONF_SCHEMA)
        self._loop = self._nyuki.loop or asyncio.get_event_loop()
        self._host = None
        self._port = None
        self._middlewares = [mw_capability]
        self._debug = False
        self._app = None
        self._handler = None
        self._server = None

    @property
    def capabilities(self):
        return self._nyuki.HTTP_RESOURCES

    def configure(self, host='0.0.0.0', port=5558, debug=False):
        self._host = host
        self._port = port
        self._debug = bool(debug)

    async def start(self):
        """
        Expose capabilities by building the HTTP server.
        The server will be started with the event loop.
        """
        self._app = web.Application(
            loop=self._loop, middlewares=self._middlewares
        )
        for resource in self._nyuki.HTTP_RESOURCES:
            resource.RESOURCE_CLASS.register(self._nyuki, self._app.router)
        log.info("Starting the http server on {}:{}".format(self._host, self._port))
        self._handler = self._app.make_handler(
            log=log, access_log=access_log, debug=self._debug
        )
        self._server = await self._loop.create_server(
            self._handler, host=self._host, port=self._port
        )

    async def stop(self):
        log.info('Stopped the http server')
        self._server.close()
        await self._handler.finish_connections()
        await self._server.wait_closed()
        self._app = None
        self._handler = None
        self._server = None



import re
from collections import namedtuple

from . import runtime


class InvalidWorkflowUri(Exception):
    pass


WorkflowUri = namedtuple(
    'WorkflowURI',
    ['template_id', 'holder', 'instance_id']
)


class URI:

    REGEX = re.compile(
        r"^nyuki://(?P<template_id>[\w-]+)"
        r"@(?P<holder>[\w-]+)"
        r"(/(?P<instance_id>[\w-]+))?$"
    )

    @staticmethod
    def instance(obj):
        tid = obj.template.uid
        try:
            iid = obj.instance.uid
        except AttributeError:
            iid = obj.uid
        return 'nyuki://{template_id}@{holder}/{instance_id}'.format(
            template_id=tid, holder=runtime.bus.name, instance_id=iid
        )

    @staticmethod
    def template(template_id):
        return 'nyuki://{template_id}@{holder}'.format(
            template_id=template_id, holder=runtime.bus.name
        )

    @classmethod
    def parse(cls, uri):
        result = re.match(cls.REGEX, uri)
        if not result:
            raise InvalidWorkflowUri(uri)
        return WorkflowUri(
            result.group('template_id'),
            result.group('holder'),
            result.group('instance_id'),
        )



def generate_factory_schema(schema=None, **rules):
    """
    Append custom object properties to the base selector schema.
    """
    if schema is None:
        schema = {'type': 'object'}
    schema['required'] = schema.get('required', []) + ['rules']
    schema['properties'] = {
        **schema.get('properties', {}),
        'rules': {'$ref': '#/definitions/rules'}
    }
    schema['definitions'] = {
        **schema.get('definitions', {}),
        **rules,
        'rules': {
            'type': 'array',
            'items': {
                'oneOf': [{'$ref': '#/definitions/condition-block'}] + [
                    {'$ref': '#/definitions/{}'.format(dtype)}
                    for dtype in rules.keys()
                ]
            }
        },
        'condition-block': {
            'type': 'object',
            'required': ['type', 'conditions'],
            'properties': {
                'type': {'type': 'string', 'enum': ['condition-block']},
                'conditions': {
                    'type': 'array',
                    'items': {
                        'oneOf': [
                            {'$ref': '#/definitions/condition-if'},
                            {'$ref': '#/definitions/condition-else'}
                        ]
                    }
                }
            }
        },
        'condition-if': {
            'type': 'object',
            'required': ['type', 'condition', 'rules'],
            'properties': {
                'type': {'type': 'string', 'enum': ['if', 'elif']},
                'condition': {'type': 'string', 'minLength': 1},
                'rules': {'$ref': '#/definitions/rules'}
            }
        },
        'condition-else': {
            'type': 'object',
            'required': ['type', 'rules'],
            'properties': {
                'type': {'type': 'string', 'enum': ['else']},
                'rules': {'$ref': '#/definitions/rules'}
            }
        }
    }
    return schema



import signal
import logging
import collections
from nyuki.api import resource, Response


log = logging.getLogger(__name__)


@resource('/samples')
class ApiSampleEmitter:

    async def get(self, request):
        if self.nyuki._sampler is None:
            return Response(status=404)
        return Response(self.nyuki._sampler.output_stats())


class StackSampler:

    """
    Basic stack sampler, inspired by https://nylas.com/blog/performance
    Can be easily paired with https://github.com/brendangregg/FlameGraph
    """

    def __init__(self, interval=0.005):
        log.info('Debug mode, sampling enabled every %s', interval)
        self.interval = interval
        self._stack_counts = collections.defaultdict(int)

    def __del__(self):
        self.stop()

    def start(self):
        signal.signal(signal.SIGVTALRM, self._sample)
        signal.setitimer(signal.ITIMER_VIRTUAL, self.interval)

    def stop(self):
        self._stack_counts = collections.defaultdict(int)
        signal.setitimer(signal.ITIMER_VIRTUAL, 0)

    def _sample(self, signum, frame):
        stack = []
        while frame is not None:
            stack.append('{}({})'.format(
                frame.f_code.co_name,
                frame.f_globals.get('__name__'),
            ))
            frame = frame.f_back

        stack = ';'.join(reversed(stack))
        self._stack_counts[stack] += 1
        signal.setitimer(signal.ITIMER_VIRTUAL, self.interval)

    def output_stats(self):
        lines = []
        ordered_stacks = sorted(
            self._stack_counts.items(),
            key=lambda kv: kv[1],
            reverse=True,
        )
        lines.extend([
            '{} {}'.format(frame, count)
            for frame, count in ordered_stacks
        ])
        return '\n'.join(lines) + '\n'



import logging
from enum import Enum
from jsonschema import ValidationError, validate as validate_schema
from jsonschema.validators import Draft4Validator, extend as extend_validator
from jsonschema._validators import type_draft4
from jsonschema import _utils

from tukio import TaskRegistry, UnknownTaskName
from tukio.workflow import WorkflowRootTaskError


log = logging.getLogger(__name__)


class ErrorInfo(Enum):
    """
    Generic error messages for template validation.
    """
    UNKNOWN = 'Unknown error'
    UNKNOWN_TASK = 'Unknown task name'
    INVALID_GRAPH = 'Invalid graph semantic'
    INVALID_TASK = 'Invalid task config'


class TemplateError(Exception):
    """
    Workflow template exception descriptor.
    """
    def __init__(self, *, key=ErrorInfo.UNKNOWN, message=None, details=None):
        self.key = key
        self.message = message or key.value
        self.details = details or list()

    def as_dict(self):
        return {
            'key': self.key.name,
            'message': self.message,
            'details': self.details
        }

    @staticmethod
    def format_details(exc, task):
        """
        Gives JSON formatted details for a given exception on a task:
        {
            "task_id": "<id>",
            "config_path": "<key.key.key>",
            "error": "<type>",
            "info": "<extra>"
        }
        """
        path = [str(i) for i in list(exc.absolute_path)]
        error = exc.validator
        value = exc.validator_value

        # Get a proper path for the `required` validation error
        if error == 'required':
            cursor = task.get('config', {})
            for key in path:
                cursor = cursor[key] if isinstance(cursor, dict) else cursor[int(key)]
            # We are looking for the missing key (not always the first item)
            for key in value:
                if key not in cursor:
                    path.append(key)
                    break

        return {
            'task_id': task['id'],
            'config_path': '.'.join(path),
            'error': error,
            'info': value
        }

    def __str__(self):
        return self.message


def validate(template):
    """
    Validate a template dict and aggregate all errors that could occur.
    Return a tempplate object or raise a ValidationError
    """
    # Task name validation
    errors = list()
    for task in template.tasks:
        try:
            TaskRegistry.get(task.name)
        except UnknownTaskName:
            errors.append(task.name)
    if errors:
        log.debug('unknown tasks: %s', errors)
        raise TemplateError(
            key=ErrorInfo.UNKNOWN_TASK,
            details=errors
        )

    # DAG semantic validation
    try:
        template.validate()
    except WorkflowRootTaskError as exc:
        log.debug('workflow validation error: %s', exc)
        raise TemplateError(
            key=ErrorInfo.INVALID_GRAPH,
            message='Workflow validation error: {}'.format(exc)
        ) from exc

    # Task config validation
    errors = list()
    tasks = template.as_dict().get('tasks', [])
    for task in tasks:
        err = validate_task(task, tasks)
        if err:
            errors.append(err)
    if errors:
        raise TemplateError(
            key=ErrorInfo.INVALID_TASK,
            details=errors
        )

    return template


def validate_task(task, tasks=None):
    """
    Validate the jsonschema configuration of a task
    """
    name = task['name']
    config = task.get('config', {})
    schema = getattr(TaskRegistry.get(name)[0], 'SCHEMA', {})
    format_checker = getattr(TaskRegistry.get(name)[0], 'FORMAT_CHECKER', None)
    try:
        validate_schema(config, schema, format_checker=format_checker)
    except ValidationError as exc:
        return TemplateError.format_details(exc, task)



from functools import singledispatch
from datetime import datetime, timedelta


@singledispatch
def serialize_object(obj):
    """
    Default serializer for python objects.
    """
    return 'Internal server data: {}'.format(type(obj))


@serialize_object.register(datetime)
def _serialize_datetime(dt):
    """
    Datetime serializer.
    """
    return dt.isoformat()



import asyncio
import logging


log = logging.getLogger(__name__)


class Service(object):

    async def start(self, *args, **kwargs):
        raise NotImplementedError

    def configure(self, *args, **kwargs):
        raise NotImplementedError

    async def stop(self, *args, **kwargs):
        raise NotImplementedError


class ServiceManager(object):

    def __init__(self, nyuki):
        self._loop = nyuki.loop or asyncio.get_event_loop()
        self._nyuki = nyuki
        self.services = dict()
        self._running = False

    def __iter__(self):
        return self.services.__iter__()

    def add(self, name, service):
        """
        Add a managed service
        """
        if not isinstance(service, Service):
            raise TypeError('service must be an instance of Service')

        self.services[name] = service
        # Run the service immetiately if the others are running
        if self._running:
            service.configure(**self._nyuki.config.get(name, {}))
            asyncio.ensure_future(service.start())

    @property
    def all(self):
        return self.services

    def get(self, name):
        return self.services[name]

    async def start(self, timeout=5):
        """
        Start all services with the given timeout
        """
        tasks = [
            asyncio.ensure_future(service.start())
            for service in self.services.values()
        ]

        log.debug('Running start tasks : %s', tasks)
        done, not_done = await asyncio.wait(tasks, timeout=timeout)
        if not_done:
            raise asyncio.TimeoutError(
                'Start tasks {} did not finish'.format(not_done)
            )
        self._running = True
        log.debug('Start tasks done')

    async def stop(self, timeout=5):
        """
        Stop all services with the given timeout
        """
        self._running = False
        tasks = [
            asyncio.ensure_future(s.stop())
            for s in self.services.values()
        ]

        done, not_done = await asyncio.wait(tasks, timeout=timeout)
        if not_done:
            log.warning('Could not stop services after %d seconds', timeout)
            log.debug('stop task not done : %s', not_done)



import json
import math
import socket
import logging
import asyncio
import aiohttp
from enum import Enum
from random import uniform

from nyuki.services import Service
from nyuki.api import Response, resource


log = logging.getLogger(__name__)


class State(Enum):
    UNKNOWN = 'unknown'
    FOLLOWER = 'follower'
    CANDIDATE = 'candidate'
    LEADER = 'leader'


class Event(Enum):
    ELECTED = 'elected'
    DISMISSED = 'dismissed'
    FAILURES = 'failures'


@resource('/raft', ['v1'], 'application/json')
class ApiRaft:
    """
    This interface enables communication between members of a Raft cluster.
    """
    async def put(self, request):
        """
        Raft candidate request.
        """
        proto = self.nyuki.raft
        # If this instance has already voted for another one
        if proto.voted_for:
            return Response(
                status=403,
                body={'voted': proto.voted_for, 'instance': proto.uid}
            )

        # Local variables
        data = await request.json()
        proto.voted_for = data['candidate']
        proto.term = data['term']

        # Reset the timer
        proto.set_timer(proto.candidate)
        return Response(status=200, body={'instance': proto.uid})

    async def post(self, request):
        """
        Heartbeat endpoint.
        """
        proto = self.nyuki.raft
        data = await request.json()
        suspicious = proto.suspicious

        # Local variables
        proto.state = State.FOLLOWER
        proto.votes = 0
        proto.voted_for = None
        proto.log = data['log']
        proto.suspicious.clear()

        # Reset the timer
        proto.set_timer(proto.candidate)
        return Response(status=200, body={
            'instance': proto.uid,
            'suspicious': list(suspicious)
        })


class RaftProtocol(Service):
    """
    Leader election based on Raft distributed algorithm.
    Paper: https://raft.github.io/raft.pdf
    """

    HEARTBEAT = 1.0
    TIMEOUT = (2.0, 3.5)

    def __init__(self, nyuki):
        self.service = nyuki.config['service']
        self.loop = nyuki.loop or asyncio.get_event_loop()
        self.uid = nyuki.id
        self.ipv4 = socket.gethostbyname(socket.gethostname())
        self.handlers = {event: set() for event in Event}

        self.cluster = {}
        self.suspicious = DanausSet(after=5, callback=self.failure_handler)
        self.timer = None
        self.state = State.UNKNOWN
        self.term = -1
        self.votes = -1
        self.voted_for = None
        self.majority = math.inf
        self.log = {}

    @property
    def network(self):
        return {**self.cluster, self.ipv4: self.uid}

    def configure(self, *args, **kwargs):
        pass

    def register(self, etype, callback):
        self.handlers[Event(etype)].add(callback)

    def set_timer(self, cb, factor=1):
        """
        Set or reset a unique timer.
        """
        if self.timer:
            self.timer.cancel()
        self.timer = self.loop.call_later(
            uniform(*self.TIMEOUT) * factor, asyncio.ensure_future, cb()
        )

    @staticmethod
    async def request(ipv4, method, data=None):
        """
        Utility method to perform HTTP requests, Raft-specific, to an instance.
        """
        request = {
            'url': 'http://{host}:5558/v1/raft'.format(host=ipv4),
            'headers': {'Content-Type': 'application/json'},
            'data': json.dumps(data or {})
        }
        try:
            async with aiohttp.ClientSession() as session:
                http_method = getattr(session, method)
                async with http_method(**request) as resp:
                    if resp.status != 200:
                        return
                    return await resp.json()
        except (aiohttp.ClientError, ConnectionError):
            return

    async def start(self, *args, **kwargs):
        """
        Starts the protocol as a follower instance.
        """
        self.state = State.FOLLOWER
        self.term = 0
        self.votes = 0
        # Won't bootstrap the timer here to avoid any unwanted early election

    async def stop(self, *args, **kwargs):
        """
        Stops the protocol by cancelling the current timer.
        """
        self.state = State.FOLLOWER
        if self.timer:
            self.timer.cancel()

    async def discovery_handler(self, addresses):
        """
        The discovery service provides updates periodically.
        """
        cluster = {ipv4: self.cluster.get(ipv4) for ipv4 in addresses}
        if self.ipv4 in cluster:
            del cluster[self.ipv4]
        else:
            log.warning("This instance isn't part of the discovery results")
            await self.stop()
            return

        # Check differences
        added = set(cluster.keys()) - set(self.cluster.keys())
        self.suspicious.update([
            (ipv4, self.cluster[ipv4] or self.log.get(ipv4))
            for ipv4 in set(self.cluster.keys()) - set(cluster.keys())
        ])
        self.cluster = cluster

        if self.state is State.LEADER:
            # Schedule HB for new workers
            for ipv4 in added:
                asyncio.ensure_future(self.heartbeat(ipv4))
        elif self.state is State.FOLLOWER and not self.timer:
            # The protocol has started but the timer needs to be bootstraped
            # Initial factor for the timer is higher (discovery reasons)
            self.set_timer(self.candidate, 5)

    async def failure_handler(self, instances):
        """
        Handle suspicous instances.
        """
        failing = [uid for ipv4, uid in instances if uid is not None]
        if not failing:
            return
        for callback in self.handlers[Event.FAILURES]:
            asyncio.ensure_future(callback(failing))

    async def candidate(self):
        """
        Election timer went out, this instance considers itself as a candidate.
        """
        cluster_size = len(self.cluster) + 1

        # Promote itself as leader if alone
        if cluster_size == 1:
            await self.promote()
            return

        # Local variables
        self.state = State.CANDIDATE
        self.term += 1
        self.votes = 1
        self.voted_for = self.uid
        self.majority = int(math.floor(cluster_size / 2) + 1)

        # Init the timer, retry vote if timeout
        log.debug("Instance is candidate (requires %d votes)", self.majority)
        self.set_timer(self.candidate)

        # Start the election
        for ipv4 in self.cluster:
            asyncio.ensure_future(self.request_vote(ipv4, self.term))

    async def promote(self):
        """
        Promote this instance to the rank of leader.
        """
        log.info("Leader elected of the service '%s'", self.service)

        # Local variables
        if self.timer:
            self.timer.cancel()
        self.state = State.LEADER
        self.votes = 0
        self.voted_for = None

        # Use the log to restore ipv4-to-uid mapping
        for ipv4, uid in self.log.items():
            if self.cluster.get(ipv4):
                self.cluster[ipv4] = uid

        # Sending heartbeats to the cluster
        for ipv4 in self.cluster:
            asyncio.ensure_future(self.heartbeat(ipv4))

    async def request_vote(self, ipv4, term):
        """
        Request a vote from an instance.
        """
        vote = await self.request(ipv4, 'put', {
            'candidate': self.uid, 'term': term
        })
        if (
            # Won't count negative feedbacks
            not vote or
            # Ignore the vote if the election is over
            self.term != term or self.loop.time() >= self.timer._when or
            # Not in a candidate anymore
            self.state is not State.CANDIDATE
        ):
            return

        # Count vote
        self.cluster[ipv4] = vote['instance']
        self.votes += 1

        # The instance becomes a leader
        if self.votes >= self.majority:
            await self.promote()

    async def heartbeat(self, ipv4):
        """
        Send a heartbeat to reset instance's timer.
        """
        if (
            # Won't send HB if the instance is not the leader anymore
            self.state is not State.LEADER or
            # Won't send HB if he instances is not in the cluster
            ipv4 not in self.cluster
        ):
            return

        # Schedule the next HB
        self.loop.call_later(
            self.HEARTBEAT, asyncio.ensure_future, self.heartbeat(ipv4)
        )

        # Heartbeats allow to refresh follower's timers and to replicate logs
        response = await self.request(ipv4, 'post', {
            'leader': self.uid,
            'log': self.network
        })

        # Empty answer or no response is suspicious
        uid = self.cluster.get(ipv4)
        if not response:
            self.suspicious.add((ipv4, uid))
            return

        # An instance isn't referenced under the same ID anymore
        if uid and uid != response['instance']:
            self.suspicious.add((ipv4, uid))
        self.cluster[ipv4] = response['instance']

        # Collect suspicious instances from heartbeat's response
        self.suspicious.update(
            [tuple(entry) for entry in response['suspicious']]
        )


class DanausSet(set):
    """
    Named after the myth of the daughters of Danaus.
    This set can be filled up but will eventually be emptied (upon timeout).
    """

    def __init__(self, seq=(), *, after=1, callback=None, loop=None):
        super().__init__(seq)
        self.loop = loop or asyncio.get_event_loop()
        self.after = after
        self.callback = callback
        self._timer = None

    def abort(self):
        if self._timer:
            self._timer.cancel()
        self._timer = None

    def add(self, item):
        if item not in self:
            self._schedule()
        super().add(item)

    def update(self, seq):
        if set(seq) - self:
            self._schedule()
        super().update(seq)

    def clear(self):
        super().clear()
        self.abort()

    def _schedule(self):
        self.abort()
        self._timer = self.loop.call_later(self.after, self._empty)

    def _empty(self):
        cleared = self.copy()
        self.clear()
        self._timer = None

        if cleared and self.callback:
            cb = self.callback
            if not asyncio.iscoroutine(self.callback):
                cb = asyncio.coroutine(cb)
            asyncio.ensure_future(cb(cleared))



from tukio.task.join import JoinTask

from .factory import FactoryTask, FACTORY_SCHEMAS
from .python_script import PythonScript
from .report import ReportTask
from .sleep import SleepTask
from .task_selector import TaskSelector
from .trigger_workflow import TriggerWorkflowTask


# Generic schema to reference a task ID
TASKID_SCHEMA = {
    'type': 'string',
    'description': 'task_id',
    'minLength': 1,
    'maxLength': 128
}


# Add a schema to the join task
JoinTask.SCHEMA = {
    'type': 'object',
    'required': ['wait_for'],
    'properties': {
        'wait_for': {
            'anyOf': [
                {
                    'type': 'array',
                    'minItems': 2,
                    'maxItems': 64,
                    'uniqueItems': True,
                    'items': TASKID_SCHEMA
                },
                {'type': 'integer', 'minimum': 2}
            ]
        },
        'timeout': {'type': 'integer', 'minimum': 1}
    }
}



import asyncio
import logging
from tukio.task import register
from tukio.task.holder import TaskHolder


log = logging.getLogger(__name__)


@register('sleep', 'execute')
class SleepTask(TaskHolder):

    """
    Mainly a testing task
    """

    SCHEMA = {
        'type': 'object',
        'properties': {
            'time': {'type': 'integer', 'minimum': 1}
        }
    }

    async def execute(self, event):
        task = asyncio.Task.current_task()
        time = self.config.get('time', 2)
        log.info('%s: sleeping %s second%s', task.uid, time, 's' if time > 1 else '')
        await asyncio.sleep(time)
        log.info('%s: done sleeping', task.uid)
        return event.data



import asyncio
import json
import logging
import re
from collections import namedtuple
from hbmqtt.client import MQTTClient, ConnectException, ClientException
from hbmqtt.errors import NoDataException
from hbmqtt.mqtt.constants import QOS_1
from uuid import uuid4

from nyuki.bus import reporting
from nyuki.services import Service
from nyuki.utils import serialize_object
from .persistence import BusPersistence, EventStatus, PersistenceError


log = logging.getLogger(__name__)


MQTTRegex = namedtuple('MQTTRegex', ['regex', 'callbacks'])


class MqttBus(Service):

    """
    Nyuki topics formatted as:
        - global publications:
            {nyuki_name}/publications
    """

    SERVICE = 'mqtt'
    CONF_SCHEMA = {
        'type': 'object',
        'required': ['bus'],
        'properties': {
            'bus': {
                'type': 'object',
                'required': ['name'],
                'properties': {
                    'cafile': {'type': 'string', 'minLength': 1},
                    'certfile': {'type': 'string', 'minLength': 1},
                    'host': {'type': 'string', 'minLength': 1},
                    'keyfile': {'type': 'string', 'minLength': 1},
                    'name': {'type': 'string', 'minLength': 1},
                    'port': {'type': 'integer'},
                    'persistence': {
                        'type': 'object',
                        'properties': {
                            'backend': {
                                'type': 'string',
                                'minLength': 1
                            }
                        }
                    },
                    'scheme': {
                        'type': 'string',
                        'enum': ['ws', 'wss', 'mqtt', 'mqtts']
                    },
                    'service': {'type': 'string', 'minLength': 1},
                    'keep_alive': {'type': 'integer', 'minimum': 1},
                    'ping_delay': {'type': 'integer', 'minimum': 1}
                },
                'additionalProperties': False
            }
        }
    }

    def __init__(self, nyuki, loop=None):
        self._nyuki = nyuki
        self._loop = loop or asyncio.get_event_loop()
        self._host = None
        self.client = None
        self._pending = {}
        self.name = None
        self._subscriptions = {}
        self._regex_subscriptions = {}

        # Coroutines
        self.connect_future = None
        self.listen_future = None

    @property
    def topics(self):
        return list(self._subscriptions.keys())

    def configure(self, name, scheme='mqtt', host='localhost', port=1883,
                  cafile=None, certfile=None, keyfile=None, persistence={},
                  service=None, keep_alive=60, ping_delay=5):
        if scheme in ['mqtts', 'wss']:
            if not cafile or not certfile or not keyfile:
                raise ValueError(
                    "secured scheme requires 'cafile', 'certfile' and 'keyfile'"
                )

        self._host = '{}://{}:{}'.format(scheme, host, port)
        self.name = name
        self._cafile = cafile
        self.client = MQTTClient(
            config={
                'auto_reconnect': False,
                'certfile': certfile,
                'keyfile': keyfile,
                'keep_alive': keep_alive,
                'ping_delay': ping_delay
            },
            loop=self._loop
        )

        # Persistence storage
        self._persistence = BusPersistence(name=name, **persistence)
        log.info('Bus persistence set to %s', self._persistence.backend)

    async def start(self):
        def cancelled(future):
            try:
                future.result()
            except asyncio.CancelledError:
                log.debug('future cancelled: %s', future)
        self.connect_future = asyncio.ensure_future(self._run())
        self.connect_future.add_done_callback(cancelled)

        try:
            await self._persistence.init()
        except PersistenceError:
            log.error(
                'Could not init persistence storage with %s',
                self._persistence.backend
            )

    async def stop(self):
        # Clean client
        if self.client is not None:
            for task in self.client.client_tasks:
                log.debug('cancelling mqtt client tasks')
                task.cancel()
            if self.client._connected_state.is_set():
                log.debug('disconnecting mqtt client')
                await self.client.disconnect()
        # Clean tasks
        if self.connect_future:
            log.debug('cancelling _run coroutine')
            self.connect_future.cancel()
        if self.listen_future:
            log.debug('cancelling _listen coroutine')
            self.listen_future.cancel()
        await self._persistence.close()
        log.info('MQTT service stopped')

    def init_reporting(self):
        """
        Initialize reporting module
        """
        reporting.init(self.name, self)

    def _regex_topic(self, topic):
        """
        Transform the MQTT pattern into a regex object.
        """
        return re.compile(r'^{}$'.format(
            topic.replace('+', '[^\/]+').replace('#', '.+')
        ))

    async def replay(self, since=None, status=None):
        """
        Replay events since the given datetime (or all if None)
        """
        msg = 'Replaying events'
        if since:
            msg += ' since {}'.format(since)
        if status:
            msg += ' with status {}'.format(status)
        log.info(msg)

        events = await self._persistence.retrieve(since, status)
        for event in events:
            # Publish events one by one in the right publish time order
            await self.publish(json.loads(
                event['message']),
                event['topic'],
                event['id']
            )

    async def subscribe(self, topic, callback):
        """
        Subscribe to a topic and setup the callback.
        This will pre-compile the regex used for this topic.
        """
        if not asyncio.iscoroutinefunction(callback):
            raise ValueError('event callback must be a coroutine')

        log.debug('MQTT subscription to %s -> %s', topic, callback.__name__)
        await self.client.subscribe([(topic, QOS_1)])

        # Regexes are about topics like 'word/+/word' or 'word/#'
        is_regex = '+' in topic or topic.endswith('#')
        if is_regex is True:
            try:
                self._regex_subscriptions[topic].callbacks.add(callback)
            except KeyError:
                self._regex_subscriptions[topic] = MQTTRegex(
                    regex=self._regex_topic(topic),
                    callbacks={callback},
                )
        # Standard topics are a simple dict/set pair.
        else:
            try:
                self._subscriptions[topic].add(callback)
            except KeyError:
                self._subscriptions[topic] = {callback}

        log.info('Subscribed to %s', topic)

    async def _unsub_regex(self, topic, callback):
        """
        Unsubscribe from a regex topic.
        """
        if topic not in self._regex_subscriptions:
            return
        if callback in self._regex_subscriptions[topic].callbacks:
            log.debug(
                'MQTT unsubscription from %s -> %s',
                topic, callback.__name__,
            )
            self._regex_subscriptions[topic].callbacks.remove(callback)
        if callback is None or not self._regex_subscriptions[topic].callbacks:
            del self._regex_subscriptions[topic]
            await self.client.unsubscribe([topic])
            log.info('Unsubscribed from %s', topic)

    async def _unsub(self, topic, callback):
        """
        Unsubscribe from a standard topic.
        """
        if topic not in self._subscriptions:
            return
        if callback in self._subscriptions[topic]:
            log.debug(
                'MQTT unsubscription from %s -> %s',
                topic, callback.__name__,
            )
            self._subscriptions[topic].remove(callback)
        if callback is None or not self._subscriptions[topic]:
            del self._subscriptions[topic]
            await self.client.unsubscribe([topic])
            log.info('Unsubscribed from %s', topic)

    async def unsubscribe(self, topic, callback=None):
        """
        Unsubscribe from a topic, remove callback if set.
        """
        if '+' in topic or topic.endswith('#'):
            await self._unsub_regex(topic, callback)
        else:
            await self._unsub(topic, callback)

    async def _resubscribe(self):
        """
        Resubscribe on reconnection.
        """
        subs = list(self._subscriptions.keys()) + \
            list(self._regex_subscriptions.keys())
        for topic in subs:
            log.debug('Resubscribing to %s', topic)
            await self.client.subscribe([(topic, QOS_1)])

    async def publish(self, data, topic=None, previous_uid=None):
        """
        Publish in given topic or default one
        """
        uid = previous_uid or str(uuid4())
        topic = topic or self.name
        log.info('Publishing an event to %s', topic)
        log.debug('dump: %s', data)
        data = json.dumps(data, default=serialize_object)

        # Store the event as PENDING if it is new
        if self._persistence and previous_uid is None:
            await self._persistence.store({
                'id': uid,
                'status': EventStatus.PENDING.value,
                'topic': topic,
                'message': data,
            })
            if self._persistence.memory_buffer.is_full:
                asyncio.ensure_future(self._nyuki.on_buffer_full(
                    self._persistence.memory_buffer.free_slot
                ))

        if self.client._connected_state.is_set():
            # Implies QOS_0
            await self.client.publish(topic, data.encode())
            status = EventStatus.SENT
            log.debug('Event successfully sent to topic %s', topic)
        else:
            status = EventStatus.FAILED
            log.error('Failed to send event to topic %s', topic)

        if self._persistence:
            await self._persistence.update(previous_uid or uid, status)

    async def _run(self):
        """
        Handle reconnection every 3 seconds
        """
        while True:
            log.info('Trying MQTT connection to %s', self._host)
            try:
                await self.client.connect(self._host, cafile=self._cafile)
            except (ConnectException, NoDataException) as exc:
                log.error(exc)
                log.info('Waiting 3 seconds to reconnect')
                await asyncio.sleep(3.0)
                continue

            # Replaying events
            log.info('Connection made with MQTT')
            if self._persistence:
                asyncio.ensure_future(self.replay(
                    status=EventStatus.not_sent()
                ))

            # Start listening
            await self._resubscribe()
            self.listen_future = asyncio.ensure_future(self._listen())
            # Blocks until mqtt is disconnected
            await self.client._handler.wait_disconnect()
            # Clean listen_future
            self.listen_future.cancel()
            self.listen_future = None

    async def _listen(self):
        """
        Listen to events after a successful connection
        """
        while True:
            try:
                message = await self.client.deliver_message()
            except ClientException as exc:
                log.error(exc)
                break

            if message is None:
                log.info('listening loop ended')
                break

            topic = message.topic
            data = json.loads(message.data.decode())

            # Iterate and call all regex topics callbacks
            for mqttregex in self._regex_subscriptions.values():
                if mqttregex.regex.match(topic):
                    log.debug("Event from topic '%s': %s", topic, data)
                    for callback in mqttregex.callbacks:
                        asyncio.ensure_future(callback(topic, data.copy()))

            try:
                # Iterate and call all single topic callbacks
                for callback in self._subscriptions[topic]:
                    asyncio.ensure_future(callback(topic, data.copy()))
            except KeyError:
                pass



from aiohttp import ClientSession
import asyncio
import json
import logging
from uuid import uuid4
from enum import Enum

from tukio.task import register
from tukio.task.holder import TaskHolder
from tukio.workflow import WorkflowExecState, Workflow

from .utils import runtime
from .utils.uri import URI


log = logging.getLogger(__name__)


class Status(Enum):
    RUNNING = 'running'
    DONE = 'done'
    TIMEOUT = 'timeout'


@register('trigger_workflow', 'execute')
class TriggerWorkflowTask(TaskHolder):

    WORKFLOW_URL = 'http://{host}{nyuki_api}/v1/workflow/instances'
    SCHEMA = {
        'type': 'object',
        'required': ['nyuki_api', 'template'],
        'properties': {
            'nyuki_api': {'type': 'string', 'description': 'nyuki_api'},
            'template': {'type': 'string', 'description': 'template_id'},
            'draft': {'type': 'boolean'},
            'await_completion': {'type': 'boolean', 'default': True},
            'timeout': {'type': 'integer', 'minimum': 1, 'default': 60}
        },
        'dependencies': {
            'await_completion': ['timeout']
        },
        'additionalProperties': False
    }

    def __init__(self, config):
        super().__init__(config)
        self.template = self.config['template']
        self.draft = self.config.get('draft', False)
        self.blocking = self.config.get('await_completion', True)
        self.timeout = self.config.get('timeout', 60)
        # Workflow URL
        self.nyuki_api = self.config.get('nyuki_api') or ''
        if self.nyuki_api and not self.nyuki_api.startswith('/'):
            self.nyuki_api = '/' + self.nyuki_api
        self.url = self.WORKFLOW_URL.format(
            host=runtime.config.get('http_host', 'localhost'),
            nyuki_api=self.nyuki_api
        )

        self._task = None
        self._triggered = {
            'workflow_holder': self.nyuki_api.split('/')[1],
            'workflow_template_id': self.template,
            'workflow_exec_id': None,
            'workflow_draft': self.draft,
            'status': None
        }

    def report(self):
        return self._triggered

    async def teardown(self):
        exec_id = self._triggered['workflow_exec_id']
        holder = self._triggered['workflow_holder']
        wf_id = '{}@{}'.format(exec_id, holder)
        if not exec_id:
            log.debug('No triggered workflow to cancel')
            return

        async with ClientSession() as session:
            params = {
                'url': '{}/{}'.format(self.url, exec_id),
                'headers': {'Content-Type': 'application/json'}
            }
            async with session.delete(**params) as response:
                response = response.status

        if response != 200:
            log.debug('Failed to cancel triggered workflow {}'.format(wf_id))
        else:
            log.debug('Triggered workflow {} has been cancelled'.format(wf_id))

    async def async_exec(self, topic, data):
        log.debug(
            "Received data for async trigger_workflow in '%s': %s",
            topic, data,
        )
        if not self.async_future.done():
            self.async_future.set_result(data)
        await runtime.bus.unsubscribe(topic)

    async def execute(self, event):
        """
        Entrypoint execution method.
        """
        data = event.data
        self._task = asyncio.Task.current_task()

        # Send the HTTP request
        log.info('Request %s to process template %s', self.nyuki_api, self.template)
        log.debug(
            'Request details: url=%s, draft=%s, data=%s',
            self.url, self.draft, data
        )

        # Setup headers (set requester and exec-track to avoid workflow loops)
        workflow = Workflow.current_workflow()
        wf_instance = runtime.workflows[workflow.uid]
        parent = wf_instance.exec.get('requester')
        track = list(wf_instance.exec.get('track', []))
        if parent:
            track.append(parent)

        headers = {
            'Content-Type': 'application/json',
            'Referer': URI.instance(workflow),
            'X-Surycat-Exec-Track': ','.join(track)
        }

        # Handle blocking trigger_workflow using mqtt
        if self.blocking:
            topic = '{}/async/{}'.format(runtime.bus.name, str(uuid4())[:8])
            headers['X-Surycat-Async-Topic'] = topic
            headers['X-Surycat-Async-Events'] = ','.join([
                WorkflowExecState.end.value,
                WorkflowExecState.error.value,
            ])
            self.async_future = asyncio.Future()
            await runtime.bus.subscribe(topic, self.async_exec)

            def _unsub(f):
                asyncio.ensure_future(runtime.bus.unsubscribe(topic))
            self._task.add_done_callback(_unsub)

        async with ClientSession() as session:
            params = {
                'url': self.url,
                'headers': headers,
                'data': json.dumps({
                    'id': self.template,
                    'draft': self.draft,
                    'inputs': data
                })
            }
            async with session.put(**params) as response:
                # Response validity
                if response.status != 200:
                    msg = "Can't process workflow template {} on {}".format(
                        self.template, self.nyuki_api
                    )
                    if response.status % 400 < 100:
                        reason = await response.json()
                        msg = "{}, reason: {}".format(msg, reason['error'])
                    raise RuntimeError(msg)
                resp_body = await response.json()

        instance = resp_body['exec']['id']
        self._triggered['workflow_exec_id'] = instance
        log.debug('Request sent successfully to {}', self.nyuki_api)

        if not self.blocking:
            self._triggered['status'] = Status.DONE.value
            self._task.dispatch_progress(self.report())

        # Block until task completed
        if self.blocking:
            self._triggered['status'] = Status.RUNNING.value
            self._task.dispatch_progress(self.report())
            log.info('Waiting for %s@%s to complete', instance, self.nyuki_api)
            try:
                await asyncio.wait_for(self.async_future, self.timeout)
                self._triggered['status'] = Status.DONE.value
                log.info('Instance %s@%s is done', instance, self.nyuki_api)
            except asyncio.TimeoutError:
                self._triggered['status'] = Status.TIMEOUT.value
                log.info('Instance %s@%s has timeouted', instance, self.nyuki_api)
            self._task.dispatch_progress({'status': self._triggered['status']})

        return data



class PersistenceBackend(object):

    """
    Base backend object for persistence, a persistence backend should
    overrides the required methods (store, retrieve).
    """

    async def init(self):
        pass

    async def ping(self):
        return True

    async def store(self, event):
        raise NotImplementedError

    async def retrieve(self, since, status):
        raise NotImplementedError



import logging
import asyncio
from tukio import Workflow
from tukio.task import register
from tukio.task.holder import TaskHolder

from nyuki.utils.evaluate import ConditionBlock
from nyuki.workflow.tasks.utils import generate_factory_schema


log = logging.getLogger(__name__)


class TaskConditionBlock(ConditionBlock):

    """
    Overrides work on ConditionBlock from the factory task to
    set next workflow tasks.
    """

    def __init__(self, conditions, workflow):
        super().__init__(conditions)
        self._workflow = workflow
        self._selected = None

    def condition_validated(self, rules, data):
        """
        Set next workflow tasks upon validating a condition.
        """
        if rules:
            self._workflow.set_next_tasks(rules[0]['tasks'])
            self._selected = rules[0]['tasks']

    def apply(self, data):
        super().apply(data)
        return self._selected


@register('task_selector', 'execute')
class TaskSelector(TaskHolder):

    SCHEMA = generate_factory_schema(tasks={
        'type': 'object',
        'required': ['type', 'tasks'],
        'properties': {
            'type': {'type': 'string', 'enum': ['task-selector']},
            'tasks': {
                'type': 'array',
                'items': {
                    'type': 'string',
                    'minLength': 1,
                    'uniqueItems': True
                }
            }
        }
    })

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._selected = None

    async def execute(self, event):
        data = event.data
        workflow = Workflow.current_workflow()
        self._selected = None
        for block in self.config['rules']:
            if block['type'] == 'task-selector':
                workflow.set_next_tasks(block['tasks'])
                self._selected = block['tasks']
            elif block['type'] == 'condition-block':
                selected = TaskConditionBlock(block['conditions'], workflow).apply(data)
                if selected is not None:
                    self._selected = selected

        task = asyncio.Task.current_task()
        task.dispatch_progress({'tasks': self._selected})
        log.debug('Tasks selected: %s', self._selected)
        return data

    def report(self):
        return {'tasks': self._selected}



import re
import logging
from pymongo.errors import AutoReconnect

from nyuki.api import Response, resource, HTTPBreak


log = logging.getLogger(__name__)


class DataInspector(object):

    REGEX = [
        re.compile('{([a-zA-Z_\-]+)}', re.IGNORECASE),
        re.compile('@([a-zA-Z_\-]+)', re.IGNORECASE)
    ]

    @classmethod
    def iter(cls, node):
        if isinstance(node, dict):
            for value in node.values():
                yield from cls.iter(value)
        elif isinstance(node, list):
            for value in node:
                yield from cls.iter(value)
        else:
            yield node

    async def required_keys(self, tid, version=None, draft=False):
        try:
            template = await self.nyuki.storage.templates.get(
                tid=tid, version=version, draft=draft
            )
        except AutoReconnect:
            raise HTTPBreak(503)
        if not template:
            raise HTTPBreak(404, {'error': 'template not found'})

        template = template[0]
        keys = set()

        for task in template.get('tasks', []):
            for key, data in task.get('config', {}).items():
                # Get all evaluable inner-data
                for value in self.iter(data):
                    if not isinstance(value, str):
                        continue
                    for regex in self.REGEX:
                        for data_key in regex.findall(value):
                            keys.add(data_key)
        return list(keys)


@resource('/workflow/vars/{tid}', versions=['v1'])
class ApiVars(DataInspector):

    async def get(self, request, tid):
        keys = await self.required_keys(tid)
        return Response(body=keys)


@resource('/workflow/vars/{tid}/{version:\d+}', versions=['v1'])
class ApiVarsVersion(DataInspector):

    async def get(self, request, tid, version):
        keys = await self.required_keys(tid, version=version)
        return Response(body=keys)


@resource('/workflow/vars/{tid}/draft', versions=['v1'])
class ApiVarsDraft(DataInspector):

    async def get(self, request, tid):
        keys = await self.required_keys(tid, draft=True)
        return Response(body=keys)



import asyncio
import json
import logging
from slixmpp import ClientXMPP
from slixmpp.exceptions import IqError, IqTimeout
from uuid import uuid4

from nyuki.bus import reporting
from nyuki.services import Service
from nyuki.utils import serialize_object

from .persistence import BusPersistence, EventStatus, PersistenceError


log = logging.getLogger(__name__)


class PublishError(Exception):
    pass


class _XmppClient(ClientXMPP):

    """
    XMPP client to connect to the bus.
    This class is based on Slixmpp (fork of Sleexmpp) using asyncio event loop.
    """

    def __init__(self, jid, password, host=None, port=5222, certificate=None):
        super().__init__(jid, password)
        host = host or self.boundjid.domain
        self._address = (host, port)

        self.register_plugin('xep_0045')  # Multi-user chat
        self.register_plugin('xep_0077')  # In-band registration

        # Disable IPv6 until we really need it
        self.use_ipv6 = False

        if certificate:
            self.ca_certs = certificate

    def connect(self, **kwargs):
        """
        Connect to the XMPP server using the default address computed at init
        if not passed as argument.
        """
        try:
            self._address = kwargs.pop('address')
        except KeyError:
            pass
        super().connect(address=self._address, **kwargs)
        self.init_plugins()

    def start_tls(self):
        try:
            return super().start_tls()
        except FileNotFoundError:
            log.error('SSL certificates missing at %s', self.ca_certs)
            self.abort()

    def exception(self, exc):
        """
        xmlstream.py catches exceptions itself, calls this method afterwards
        """
        reporting.exception(exc)


class XmppBus(Service):

    """
    A simple class that implements Publish/Subscribe-like communications over
    XMPP. This communication layer is also called "bus".
    Each nyuki has its own "topic" (actually a MUC) to publish events. And it
    can publish events to that topic only. In the meantime, each nyuki can
    subscribe to any other topics to process events from other nyukis.
    """

    SERVICE = 'xmpp'
    CONF_SCHEMA = {
        "type": "object",
        "required": ["bus"],
        "properties": {
            "bus": {
                "type": "object",
                "required": ["jid", "password"],
                "properties": {
                    "certificate": {"type": "string", "minLength": 1},
                    "host": {"type": "string", "minLength": 1},
                    "jid": {"type": "string", "minLength": 1},
                    "muc_domain": {"type": ["string", "null"]},
                    "password": {"type": "string", "minLength": 1},
                    "persistence": {
                        "type": "object",
                        "properties": {
                            "backend": {
                                "type": "string",
                                "minLength": 1
                            }
                        }
                    },
                    "port": {"type": "integer"},
                    "service": {"type": "string", "minLength": 1}
                },
                "additionalProperties": False
            }
        }
    }

    def __init__(self, nyuki):
        self._nyuki = nyuki
        self._nyuki.register_schema(self.CONF_SCHEMA)
        self._loop = self._nyuki.loop or asyncio.get_event_loop()

        self.client = None

        # Connectivity stuff
        self.reconnect = False
        self._connected = asyncio.Event()
        self._persistence = None
        self._publish_futures = dict()

        # MUCs
        self.name = None
        self._callbacks = dict()
        self._mucs = None
        self._muc_domain = None

    @property
    def topics(self):
        if not self._mucs:
            return []
        return [topic.split('@')[0] for topic in self._mucs.rooms.keys()]

    async def start(self, timeout=0):
        self.client.connect()
        if timeout:
            await asyncio.wait_for(self._connected.wait(), timeout)

        try:
            await self._persistence.init()
        except PersistenceError:
            log.error(
                'Could not init persistence storage with %s',
                self._persistence.backend
            )

    def configure(self, jid, password, host='localhost', port=5222,
                  muc_domain='mucs.localhost', certificate=None,
                  persistence={}, service=None):
        # XMPP client and main handlers
        self.client = _XmppClient(
            jid, password, host, port, certificate=certificate
        )
        self.client.loop = self._loop
        self.client.add_event_handler('connection_failed', self._on_failure)
        self.client.add_event_handler('disconnected', self._on_disconnect)
        self.client.add_event_handler('message', self._on_direct_message)
        self.client.add_event_handler('register', self._on_register)
        self.client.add_event_handler('session_start', self._on_start)
        self.client.add_event_handler('stream_error', self._on_stream_error)

        # MUCs
        self.name = self.client.boundjid.user
        self._callbacks[self.name] = None
        self._mucs = self.client.plugin['xep_0045']
        self._muc_domain = muc_domain
        self.client.add_event_handler('groupchat_message', self._on_event)
        self.client.add_event_handler(
            'muc::{}::message_error'.format(self._muc_address(self.name)),
            self._on_failed_publish
        )

        # Persistence storage
        self._persistence = BusPersistence(name=self.name, **persistence)
        log.info('Bus persistence set to %s', self._persistence.backend)

    async def stop(self):
        if not self.client:
            return

        if not self.client.transport:
            log.warning('XMPP client is already disconnected')
            return

        self.reconnect = False
        self.client.disconnect(wait=2)

        try:
            await asyncio.wait_for(self.client.disconnected, 2.0)
        except asyncio.TimeoutError:
            log.error('Could not end bus connection after 2 seconds')

        await self._persistence.close()

    def init_reporting(self):
        """
        Initialize reporting module
        """
        reporting.init(self.client.boundjid.user, self)

    def _muc_address(self, topic):
        return '{}@{}'.format(topic, self._muc_domain)

    async def _on_register(self, event):
        """
        XMPP event handler while registering a user (in-band registration).
        """
        resp = self.client.Iq()
        resp['type'] = 'set'
        resp['register']['username'] = self.client.boundjid.user
        resp['register']['password'] = self.client.password

        try:
            await resp.send()
        except IqError as exc:
            error = exc.iq['error']['text']
            log.debug("Could not register account: {}".format(error))
        except IqTimeout:
            log.error("No response from the server")
        else:
            log.debug("Account {} created".format(self.client.boundjid))

    async def _on_start(self, event):
        """
        XMPP event handler when the connection has been made.
        """
        log.info('Connected to XMPP server at {}:{}'.format(
            *self.client._address
        ))
        self.reconnect = True
        self.client.send_presence()
        self.client.get_roster()

        # Auto-subscribe to the topic where the nyuki could publish events.
        self._connected.set()
        # Reconnect to all channels (+ `self.name`, set up in configure())
        await self._resubscribe()

        # Replay events that have been lost, if any
        if self._persistence:
            await self.replay(status=EventStatus.FAILED, wait=3)

    async def _on_disconnect(self, event):
        """
        XMPP event handler when the client has been disconnected.
        """
        if self._connected.is_set():
            log.warning('Disconnected from the bus')
            self._connected.clear()

        for future in self._publish_futures.values():
            future.cancel()

        if self.reconnect:
            # Restart the connection loop
            await self.client._connect_routine()

    def _on_failure(self, event):
        """
        XMPP event handler when something is going wrong with the connection.
        """
        log.error('Connection to XMPP server {}:{} failed'.format(
            *self.client._address
        ))
        self.client.abort()

    async def _on_stream_error(self, event):
        """
        Received on stream_error from prosody
        (TODO: event informations unknown)
        """
        log.error('Received unexpected stream error, resubscribing')
        log.debug('event dump: %s', event)
        await self._resubscribe()
        if event['id'] in self._publish_futures:
            if not self._publish_futures[event['id']].done():
                self._publish_futures[event['id']].set_exception(PublishError)
            else:
                log.debug('Publish future was already done')

    async def _on_failed_publish(self, event):
        """
        Event received when a publish has failed
        (publish informations in the event)
        """
        uid = event['id']
        muc = event['from'].user
        log.warning("Was not subscribed to MUC '%s', retrying", muc)
        await self.subscribe(muc, self._callbacks.get(muc))
        if not self._publish_futures[uid].done():
            self._publish_futures[uid].set_exception(PublishError)
        else:
            log.debug('Publish future was already done for id %s', uid)

    async def _on_event(self, event):
        """
        XMPP event handler when a Nyuki Event has been received.
        """
        # Ignore events from the nyuki itself
        # Event's resource is checked as well in case of self monitoring:
        # 'monitoring@mucs.localhost/this_nyuki'
        efrom = event['from'].user
        if efrom == self.name or event['from'].resource == self.name:
            future = self._publish_futures.get(event['id'])
            if not future:
                log.warning('Received own publish that was not in memory')
            else:
                future.set_result(None)
            return

        log.debug('event received: {}'.format(event))

        try:
            body = json.loads(event['body'])
        except ValueError:
            body = {}

        callback = self._callbacks.get(efrom)
        if callback:
            log.debug('calling callback %s', callback)
            if not asyncio.iscoroutinefunction(callback):
                log.warning('event callbacks must be coroutines')
                callback = asyncio.coroutine(callback)
            await callback(efrom, body)
        else:
            log.warning('No callback set for event from %s', efrom)

    async def replay(self, since=None, status=None, wait=0):
        """
        Replay events since the given datetime (or all if None)
        """
        log.info('Replaying events')
        if since:
            log.info('    since %s', since)
        if status:
            log.info('    with status %s', status)

        for future in self._publish_futures.values():
            future.cancel()

        if wait:
            log.info('Waiting %d seconds', wait)
            await asyncio.sleep(wait)

        events = await self._persistence.retrieve(since, status)
        for event in events:
            # Publish events one by one in the right publish time order
            await self.publish(
                json.loads(event['message']),
                topic=event['topic'],
                previous_uid=event['id']
            )

    async def publish(self, event, topic=None, previous_uid=None):
        """
        Send an event in the nyuki's own MUC so that other nyukis that joined
        the MUC can process it.
        """
        if not isinstance(event, dict):
            raise TypeError('Message must be a dict')
        if self._muc_domain is None:
            log.error('No subscription to any muc')
            return
        if topic is not None:
            # XMPP does not handle MUCs containing '/'
            topic = topic.replace('/', '.')
        if topic is not None and topic not in self.topics:
            # Automatically subscribe if required
            await self.subscribe(topic)

        msg = self.client.Message()
        msg['id'] = uid = previous_uid or str(uuid4())
        msg['type'] = 'groupchat'
        msg['to'] = self._muc_address(topic or self.name)
        msg['body'] = json.dumps(event, default=serialize_object)

        self._publish_futures[uid] = asyncio.Future()
        status = EventStatus.PENDING

        # Store the event as PENDING if it is new
        if self._persistence and previous_uid is None:
            await self._persistence.store({
                'id': uid,
                'status': status.value,
                'topic': topic or self.name,
                'message': msg['body'],
            })
            in_memory = self._persistence.memory_buffer
            if in_memory.is_full:
                asyncio.ensure_future(self._nyuki.on_buffer_full(
                    in_memory.free_slot
                ))

        # Publish in MUC
        if self._connected.is_set():
            log.debug(">> publishing to '{}': {}".format(topic, event))
            log.info('Publishing an event to %s', msg['to'])
            status = None
            while True:
                msg.send()
                try:
                    await asyncio.wait_for(self._publish_futures[uid], 10.0)
                except asyncio.TimeoutError:
                    log.warning('Publication timed out, disconnecting slixmpp')
                    status = EventStatus.FAILED
                    # Crash connection and try to reconnect
                    self.reconnect = True
                    self.client.abort()
                    break
                except asyncio.CancelledError:
                    log.warning('Publication cancelled due to disconnection')
                    status = EventStatus.FAILED
                    break
                except PublishError:
                    status = EventStatus.FAILED
                    self._publish_futures[uid] = asyncio.Future()
                else:
                    log.info("Event successfully sent to MUC '%s'", msg['to'])
                    status = EventStatus.SENT
                    break
        else:
            status = EventStatus.FAILED

        del self._publish_futures[uid]
        # Once we have a result, update the stored event
        if self._persistence:
            await self._persistence.update(previous_uid or uid, status)

    async def _resubscribe(self):
        """
        Resubscribe everywhere
        """
        for topic in self._callbacks.keys():
            await self.subscribe(topic, self._callbacks.get(topic))

    async def subscribe(self, topic, callback=None):
        """
        Enter into a new chatroom using xep_0045.
        Room address format must be like '{topic}@mucs.localhost'
        """
        if not self._connected.is_set():
            log.warning("Waiting for a connection to subscribe to '%s'", topic)
        await self._connected.wait()

        # '/' are not supported in MUCs name
        topic = topic.replace('/', '.')
        self._mucs.join_muc(self._muc_address(topic), self.name)
        log.info("Subscribed to '%s'", topic)
        if self._callbacks.get(topic) is None:
            self._callbacks[topic] = callback
        else:
            log.warning("Callback already set for topic: %s", topic)

    async def unsubscribe(self, topic):
        """
        Leave a chatroom.
        """
        if not self._connected.is_set():
            log.warning("Waiting for a connection to unsubscribe from '%s'", topic)
        await self._connected.wait()

        self._mucs.leaveMUC(self._muc_address(topic), self.name)
        if topic in self._callbacks:
            del self._callbacks[topic]
        log.info("Unsubscribed from '%s'", topic)

    def direct_subscribe(self, callback):
        """
        Enable handling of direct messages from other nyukis
        """
        if not asyncio.iscoroutinefunction(callback):
            log.warning('event callbacks must be coroutines')
            callback = asyncio.coroutine(callback)
        self._direct_message_callback = callback
        log.info("Subscribed to direct messages")

    def direct_unsubscribe(self, source):
        """
        Remove the callback for direct messages
        """
        self._direct_message_callback = None
        log.info('Unsubscribed from direct messages')

    async def _on_direct_message(self, event):
        """
        Direct message events end up here
        """
        if event.get('type') == 'groupchat':
            return

        log.debug('direct message received: {}'.format(event))
        efrom = event['from'].user

        try:
            body = json.loads(event['body'])
        except ValueError:
            body = {}

        if self._direct_message_callback:
            log.debug('calling direct message callback')
            await self._direct_message_callback(efrom, body)
        else:
            log.warning('No callback set for direct messages')

    async def send_message(self, recipient, data):
        """
        Send a direct message to 'recipient'
        """
        if not recipient:
            log.debug('No recipient for direct message, ignoring it')
            return
        if not isinstance(data, dict):
            raise TypeError('Message must be a dict')
        if not self._connected.is_set():
            log.info("Waiting for a connection to direct message '%s'", recipient)
        await self._connected.wait()
        log.debug(">> direct message to '{}': {}".format(recipient, data))
        body = json.dumps(data)
        self.client.send_message(recipient, body)



import asyncio

from nyuki.bus.persistence import EventStatus
from nyuki.utils import from_isoformat

from .api import Response, resource


@resource('/bus/replay', versions=['v1'])
class ApiBusReplay:

    async def post(self, request):
        body = await request.json()

        try:
            self.nyuki._services.get('bus')
        except KeyError:
            return Response(status=404)

        # Format 'since' parameter from isoformat
        since = body.get('since')
        if since:
            try:
                since = from_isoformat(since)
            except ValueError:
                return Response({
                    'error': 'Unknown datetime format: %s'.format(since)
                }, status=400)

        # Check and parse event status
        request_status = body.get('status')
        status = list()
        if request_status:
            try:
                if isinstance(request_status, list):
                    for es in request_status:
                        status.append(EventStatus[es])
                else:
                    status.append(EventStatus[request_status])
            except KeyError:
                return Response(status=400, body={
                    'error': 'unknown event status type {}'.format(es)
                })

        await self.nyuki.bus.replay(since, status)


@resource('/bus/topics', versions=['v1'])
class ApiBusTopics:

    async def get(self, request):
        try:
            self.nyuki._services.get('bus')
        except KeyError:
            return Response(status=404)
        return Response(self.nyuki.bus.topics)


@resource('/bus/publish', versions=['v1'])
class ApiBusPublish:

    async def post(self, request):
        try:
            self.nyuki._services.get('bus')
        except KeyError:
            return Response(status=404)
        request = await request.json()
        asyncio.ensure_future(self.nyuki.bus.publish(
            request.get('data', {}), request.get('topic')
        ))



import sys


class RuntimeContext(object):

    """
    Handle runtime context as a singleton.
    """

    _instance = None

    @classmethod
    def instance(cls):
        if not cls._instance:
            cls._instance = cls()
        return cls._instance

    def __init__(self):
        self._config = dict()
        self._bus = None

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, value):
        self._config = value

    @property
    def bus(self):
        return self._bus

    @bus.setter
    def bus(self, value):
        self._bus = value


sys.modules[__name__] = RuntimeContext.instance()



import asyncio
import logging
from aiodns import DNSResolver
from aiodns.error import DNSError

from nyuki.discovery import DiscoveryService


log = logging.getLogger(__name__)


class DnsDiscovery(DiscoveryService):

    SCHEME = 'dns'
    CONF_SCHEMA = {
        "type": "object",
        "properties": {
            "discovery": {
                "type": "object",
                "properties": {
                    "method": {"type": "string", "enum": ["dns"]},
                    "entry": {"type": "string", "minLength": 1},
                    "period": {"type": "integer", "minimum": 1}
                },
                "additionalProperties": False
            }
        }
    }

    _RETRY_PERIOD = 5

    def __init__(self, nyuki, loop=None):
        self._nyuki = nyuki
        self._entry = None
        self._period = None
        self._future = None
        self._callbacks = []
        self._resolver = DNSResolver(loop=loop or asyncio.get_event_loop())

        self._nyuki.register_schema(self.CONF_SCHEMA)

    def configure(self, entry=None, period=2, **kwargs):
        self._entry = entry or self._nyuki.config['service']
        self._period = period

    def register(self, callback):
        if not callable(callback) or callback in self._callbacks:
            raise ValueError('Invalid or already registered callback')
        self._callbacks.append(callback)

    async def start(self, *args, **kwargs):
        self._future = asyncio.ensure_future(self.periodic_query())

    async def periodic_query(self):
        while True:
            try:
                answers = await self._resolver.query(self._entry, 'A')
            except DNSError as exc:
                log.error("DNS query failed for discovery service")
                log.debug("DNS failure reason: %s", str(exc))
                await asyncio.sleep(self._RETRY_PERIOD)
                continue
            addresses = [record.host for record in answers]

            # Trigger callbacks for discovered instances IPs
            for callback in self._callbacks:
                coro = (
                    callback if asyncio.iscoroutinefunction(callback)
                    else asyncio.coroutine(callback)
                )
                asyncio.ensure_future(coro(addresses))

            # Periodically execute this method
            await asyncio.sleep(self._period)

    async def stop(self):
        self._future.cancel()
